# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)


from spnego._sspi_raw.sspi import (
    accept_security_context,
    acquire_credentials_handle,
    ClientContextAttr,
    ClientContextReq,
    Credential,
    CredentialUse,
    decrypt_message,
    encrypt_message,
    initialize_security_context,
    make_signature,
    query_context_attributes,
    SecBuffer,
    SecBufferDesc,
    SecBufferType,
    SecPkgAttr,
    SecPkgAttrSizes,
    SecPkgInfo,
    SecStatus,
    SecurityContext,
    ServerContextAttr,
    ServerContextReq,
    SSPIQoP,
    TargetDataRep,
    verify_signature,
    WinNTAuthIdentity,
)



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from collections import (
    namedtuple,
)

from spnego._text import (
    to_native,
)

from cpython.exc cimport (
    PyErr_SetFromWindowsErr,
)

from libc.stdlib cimport (
    calloc,
    free,
    malloc,
)

from libc.string cimport (
    memcpy,
)

from spnego._sspi_raw.security cimport (
    AcceptSecurityContext,
    AcquireCredentialsHandleW,
    ASC_REQ_DELEGATE,
    ASC_REQ_MUTUAL_AUTH,
    ASC_REQ_REPLAY_DETECT,
    ASC_REQ_SEQUENCE_DETECT,
    ASC_REQ_CONFIDENTIALITY,
    ASC_REQ_USE_SESSION_KEY,
    ASC_REQ_ALLOCATE_MEMORY,
    ASC_REQ_USE_DCE_STYLE,
    ASC_REQ_DATAGRAM,
    ASC_REQ_CONNECTION,
    ASC_REQ_CALL_LEVEL,
    ASC_REQ_FRAGMENT_SUPPLIED,
    ASC_REQ_EXTENDED_ERROR,
    ASC_REQ_STREAM,
    ASC_REQ_INTEGRITY,
    ASC_REQ_LICENSING,
    ASC_REQ_IDENTIFY,
    ASC_REQ_ALLOW_NULL_SESSION,
    ASC_REQ_ALLOW_NON_USER_LOGONS,
    ASC_REQ_ALLOW_CONTEXT_REPLAY,
    ASC_REQ_FRAGMENT_TO_FIT,
    ASC_REQ_NO_TOKEN,
    ASC_REQ_PROXY_BINDINGS,
    ASC_REQ_ALLOW_MISSING_BINDINGS,
    ASC_RET_DELEGATE,
    ASC_RET_MUTUAL_AUTH,
    ASC_RET_REPLAY_DETECT,
    ASC_RET_SEQUENCE_DETECT,
    ASC_RET_CONFIDENTIALITY,
    ASC_RET_USE_SESSION_KEY,
    ASC_RET_ALLOCATED_MEMORY,
    ASC_RET_USED_DCE_STYLE,
    ASC_RET_DATAGRAM,
    ASC_RET_CONNECTION,
    ASC_RET_CALL_LEVEL,
    ASC_RET_THIRD_LEG_FAILED,
    ASC_RET_EXTENDED_ERROR,
    ASC_RET_STREAM,
    ASC_RET_INTEGRITY,
    ASC_RET_LICENSING,
    ASC_RET_IDENTIFY,
    ASC_RET_NULL_SESSION,
    ASC_RET_ALLOW_NON_USER_LOGONS,
    ASC_RET_ALLOW_CONTEXT_REPLAY,
    ASC_RET_FRAGMENT_ONLY,
    ASC_RET_NO_TOKEN,
    ASC_RET_NO_ADDITIONAL_TOKEN,
    CompleteAuthToken,
    CredHandle,
    CtxtHandle,
    DecryptMessage,
    DeleteSecurityContext,
    EncryptMessage,
    FreeContextBuffer,
    FreeCredentialsHandle,
    InitializeSecurityContextW,
    ISC_REQ_ALLOCATE_MEMORY,
    ISC_REQ_CONFIDENTIALITY,
    ISC_REQ_CONNECTION,
    ISC_REQ_DELEGATE,
    ISC_REQ_EXTENDED_ERROR,
    ISC_REQ_IDENTIFY,
    ISC_REQ_INTEGRITY,
    ISC_REQ_MANUAL_CRED_VALIDATION,
    ISC_REQ_MUTUAL_AUTH,
    ISC_REQ_NO_INTEGRITY,
    ISC_REQ_REPLAY_DETECT,
    ISC_REQ_SEQUENCE_DETECT,
    ISC_REQ_STREAM,
    ISC_REQ_USE_SESSION_KEY,
    ISC_REQ_USE_SUPPLIED_CREDS,
    ISC_RET_DELEGATE,
    ISC_RET_MUTUAL_AUTH,
    ISC_RET_REPLAY_DETECT,
    ISC_RET_SEQUENCE_DETECT,
    ISC_RET_CONFIDENTIALITY,
    ISC_RET_USE_SESSION_KEY,
    ISC_RET_USED_COLLECTED_CREDS,
    ISC_RET_USED_SUPPLIED_CREDS,
    ISC_RET_ALLOCATED_MEMORY,
    ISC_RET_USED_DCE_STYLE,
    ISC_RET_DATAGRAM,
    ISC_RET_CONNECTION,
    ISC_RET_INTERMEDIATE_RETURN,
    ISC_RET_CALL_LEVEL,
    ISC_RET_EXTENDED_ERROR,
    ISC_RET_STREAM,
    ISC_RET_INTEGRITY,
    ISC_RET_IDENTIFY,
    ISC_RET_NULL_SESSION,
    ISC_RET_MANUAL_CRED_VALIDATION,
    ISC_RET_FRAGMENT_ONLY,
    ISC_RET_FORWARD_CREDENTIALS,
    ISC_RET_USED_HTTP_STYLE,
    ISC_RET_NO_ADDITIONAL_TOKEN,
    ISC_RET_REAUTHENTICATION,
    MakeSignature,
    PCtxtHandle,
    PSEC_WINNT_AUTH_IDENTITY_W,
    PSecBuffer,
    PSecPkgInfoW,
    QueryContextAttributesW,
    SEC_WINNT_AUTH_IDENTITY_UNICODE,
    SEC_WINNT_AUTH_IDENTITY_W,
    SecBuffer as NativeSecBuffer,
    SecBufferDesc as NativeSecBufferDesc,
    SECBUFFER_VERSION,
    SECBUFFER_EMPTY,
    SECBUFFER_DATA,
    SECBUFFER_TOKEN,
    SECBUFFER_PKG_PARAMS,
    SECBUFFER_MISSING,
    SECBUFFER_EXTRA,
    SECBUFFER_STREAM_TRAILER,
    SECBUFFER_STREAM_HEADER,
    SECBUFFER_NEGOTIATION_INFO,
    SECBUFFER_PADDING,
    SECBUFFER_STREAM,
    SECBUFFER_MECHLIST,
    SECBUFFER_MECHLIST_SIGNATURE,
    SECBUFFER_TARGET,
    SECBUFFER_CHANNEL_BINDINGS,
    SECBUFFER_CHANGE_PASS_RESPONSE,
    SECBUFFER_TARGET_HOST,
    SECBUFFER_ATTRMASK,
    SECBUFFER_READONLY,
    SECBUFFER_READONLY_WITH_CHECKSUM,
    SECBUFFER_RESERVED,
    SecPkgContext_Names,
    SecPkgContext_PackageInfoW,
    SecPkgContext_SessionKey,
    SecPkgContext_Sizes,
    SecPkgInfoW,
    SECPKG_ATTR_NAMES,
    SECPKG_ATTR_PACKAGE_INFO,
    SECPKG_ATTR_SESSION_KEY,
    SECPKG_ATTR_SIZES,
    SECPKG_CRED_BOTH,
    SECPKG_CRED_INBOUND,
    SECPKG_CRED_OUTBOUND,
    SECQOP_WRAP_NO_ENCRYPT,
    SECURITY_INTEGER,
    SECURITY_NATIVE_DREP,
    SECURITY_NETWORK_DREP,
    SEC_E_BUFFER_TOO_SMALL as _SEC_E_BUFFER_TOO_SMALL,
    SEC_E_CONTEXT_EXPIRED as _SEC_E_CONTEXT_EXPIRED,
    SEC_E_CRYPTO_SYSTEM_INVALID as _SEC_E_CRYPTO_SYSTEM_INVALID,
    SEC_E_INCOMPLETE_MESSAGE as _SEC_E_INCOMPLETE_MESSAGE,
    SEC_E_INSUFFICIENT_MEMORY as _SEC_E_INSUFFICIENT_MEMORY,
    SEC_E_INTERNAL_ERROR as _SEC_E_INTERNAL_ERROR,
    SEC_E_INVALID_HANDLE as _SEC_E_INVALID_HANDLE,
    SEC_E_INVALID_TOKEN as _SEC_E_INVALID_TOKEN,
    SEC_E_LOGON_DENIED as _SEC_E_LOGON_DENIED,
    SEC_E_MESSAGE_ALTERED as _SEC_E_MESSAGE_ALTERED,
    SEC_E_NOT_OWNER as _SEC_E_NOT_OWNER,
    SEC_E_NO_AUTHENTICATING_AUTHORITY as _SEC_E_NO_AUTHENTICATING_AUTHORITY,
    SEC_E_NO_CREDENTIALS as _SEC_E_NO_CREDENTIALS,
    SEC_E_OK as _SEC_E_OK,
    SEC_E_OUT_OF_SEQUENCE as _SEC_E_OUT_OF_SEQUENCE,
    SEC_E_QOP_NOT_SUPPORTED as _SEC_E_QOP_NOT_SUPPORTED,
    SEC_E_SECPKG_NOT_FOUND as _SEC_E_SECPKG_NOT_FOUND,
    SEC_E_TARGET_UNKNOWN as _SEC_E_TARGET_UNKNOWN,
    SEC_E_UNKNOWN_CREDENTIALS as _SEC_E_UNKNOWN_CREDENTIALS,
    SEC_E_UNSUPPORTED_FUNCTION as _SEC_E_UNSUPPORTED_FUNCTION,
    SEC_E_WRONG_PRINCIPAL as _SEC_E_WRONG_PRINCIPAL,
    SEC_I_COMPLETE_AND_CONTINUE as _SEC_I_COMPLETE_AND_CONTINUE,
    SEC_I_COMPLETE_NEEDED as _SEC_I_COMPLETE_NEEDED,
    SEC_I_CONTINUE_NEEDED as _SEC_I_CONTINUE_NEEDED,
    SEC_I_INCOMPLETE_CREDENTIALS as _SEC_I_INCOMPLETE_CREDENTIALS,
    VerifySignature,
)

from spnego._sspi_raw.text cimport (
    u16_to_text,
    WideChar,
)


class ClientContextAttr:
    delegate = ISC_RET_DELEGATE
    mutual_auth = ISC_RET_MUTUAL_AUTH
    replay_detect = ISC_RET_REPLAY_DETECT
    sequence_detect = ISC_RET_SEQUENCE_DETECT
    confidentiality = ISC_RET_CONFIDENTIALITY
    use_session_key = ISC_RET_USE_SESSION_KEY
    used_collected_creds = ISC_RET_USED_COLLECTED_CREDS
    used_supplied_creds = ISC_RET_USED_SUPPLIED_CREDS
    allocated_memory = ISC_RET_ALLOCATED_MEMORY
    used_dce_style = ISC_RET_USED_DCE_STYLE
    datagram = ISC_RET_DATAGRAM
    connection = ISC_RET_CONNECTION
    intermediate_return = ISC_RET_INTERMEDIATE_RETURN
    call_level = ISC_RET_CALL_LEVEL
    extended_error = ISC_RET_EXTENDED_ERROR
    stream = ISC_RET_STREAM
    integrity = ISC_RET_INTEGRITY
    identify = ISC_RET_IDENTIFY
    null_session = ISC_RET_NULL_SESSION
    manual_cred_validation = ISC_RET_MANUAL_CRED_VALIDATION
    fragment_only = ISC_RET_FRAGMENT_ONLY
    forward_credentials = ISC_RET_FORWARD_CREDENTIALS
    used_http_style = ISC_RET_USED_HTTP_STYLE
    no_additional_token = ISC_RET_NO_ADDITIONAL_TOKEN
    reauthentication = ISC_RET_REAUTHENTICATION


class ClientContextReq:
    allocate_memory = ISC_REQ_ALLOCATE_MEMORY
    confidentiality = ISC_REQ_CONFIDENTIALITY
    connection = ISC_REQ_CONNECTION
    delegate = ISC_REQ_DELEGATE
    extended_error = ISC_REQ_EXTENDED_ERROR
    http = 0x10000000  # ISC_REQ_HTTP
    integrity = ISC_REQ_INTEGRITY
    identify = ISC_REQ_IDENTIFY
    manual_cred_validation = ISC_REQ_MANUAL_CRED_VALIDATION
    mutual_auth = ISC_REQ_MUTUAL_AUTH
    no_integrity = ISC_REQ_NO_INTEGRITY
    replay_detect = ISC_REQ_REPLAY_DETECT
    sequence_detect = ISC_REQ_SEQUENCE_DETECT
    stream = ISC_REQ_STREAM
    use_session_key = ISC_REQ_USE_SESSION_KEY
    use_supplied_creds = ISC_REQ_USE_SUPPLIED_CREDS


class CredentialUse:
    inbound = SECPKG_CRED_INBOUND
    outbound = SECPKG_CRED_OUTBOUND
    both = SECPKG_CRED_BOTH


class SecBufferType:
    empty = SECBUFFER_EMPTY
    data = SECBUFFER_DATA
    token = SECBUFFER_TOKEN
    pkg_params = SECBUFFER_PKG_PARAMS
    missing = SECBUFFER_MISSING
    extra = SECBUFFER_EXTRA
    stream_trailer = SECBUFFER_STREAM_TRAILER
    stream_header = SECBUFFER_STREAM_HEADER
    negotiation_info = SECBUFFER_NEGOTIATION_INFO
    padding = SECBUFFER_PADDING
    stream = SECBUFFER_STREAM
    mechlist = SECBUFFER_MECHLIST
    mechlist_signature = SECBUFFER_MECHLIST_SIGNATURE
    target = SECBUFFER_TARGET
    channel_bindings = SECBUFFER_CHANNEL_BINDINGS
    change_pass_response = SECBUFFER_CHANGE_PASS_RESPONSE
    target_host = SECBUFFER_TARGET_HOST
    attrmask = SECBUFFER_ATTRMASK
    readonly = SECBUFFER_READONLY
    readonly_with_checksum = SECBUFFER_READONLY_WITH_CHECKSUM
    reserved = SECBUFFER_RESERVED


class SecPkgAttr:
    names = SECPKG_ATTR_NAMES
    package_info = SECPKG_ATTR_PACKAGE_INFO
    session_key = SECPKG_ATTR_SESSION_KEY
    sizes = SECPKG_ATTR_SIZES


class SecStatus:
    SEC_E_BUFFER_TOO_SMALL = _SEC_E_BUFFER_TOO_SMALL
    SEC_E_CONTEXT_EXPIRED = _SEC_E_CONTEXT_EXPIRED
    SEC_E_CRYPTO_SYSTEM_INVALID = _SEC_E_CRYPTO_SYSTEM_INVALID
    SEC_E_INCOMPLETE_MESSAGE = _SEC_E_INCOMPLETE_MESSAGE
    SEC_E_INSUFFICIENT_MEMORY = _SEC_E_INSUFFICIENT_MEMORY
    SEC_E_INTERNAL_ERROR = _SEC_E_INTERNAL_ERROR
    SEC_E_INVALID_HANDLE = _SEC_E_INVALID_HANDLE
    SEC_E_INVALID_TOKEN = _SEC_E_INVALID_TOKEN
    SEC_E_LOGON_DENIED = _SEC_E_LOGON_DENIED
    SEC_E_MESSAGE_ALTERED = _SEC_E_MESSAGE_ALTERED
    SEC_E_NOT_OWNER = _SEC_E_NOT_OWNER
    SEC_E_NO_AUTHENTICATING_AUTHORITY = _SEC_E_NO_AUTHENTICATING_AUTHORITY
    SEC_E_NO_CREDENTIALS = _SEC_E_NO_CREDENTIALS
    SEC_E_OK = _SEC_E_OK
    SEC_E_OUT_OF_SEQUENCE = _SEC_E_OUT_OF_SEQUENCE
    SEC_E_QOP_NOT_SUPPORTED = _SEC_E_QOP_NOT_SUPPORTED
    SEC_E_SECPKG_NOT_FOUND = _SEC_E_SECPKG_NOT_FOUND
    SEC_E_TARGET_UNKNOWN = _SEC_E_TARGET_UNKNOWN
    SEC_E_UNKNOWN_CREDENTIALS = _SEC_E_UNKNOWN_CREDENTIALS
    SEC_E_UNSUPPORTED_FUNCTION = _SEC_E_UNSUPPORTED_FUNCTION
    SEC_E_WRONG_PRINCIPAL = _SEC_E_WRONG_PRINCIPAL
    SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
    SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
    SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED
    SEC_I_INCOMPLETE_CREDENTIALS = _SEC_I_INCOMPLETE_CREDENTIALS


class ServerContextAttr:
    delegate = ASC_RET_DELEGATE
    mutual_auth = ASC_RET_MUTUAL_AUTH
    replay_detect = ASC_RET_REPLAY_DETECT
    sequence_detect = ASC_RET_SEQUENCE_DETECT
    confidentiality = ASC_RET_CONFIDENTIALITY
    use_session_key = ASC_RET_USE_SESSION_KEY
    allocated_memory = ASC_RET_ALLOCATED_MEMORY
    used_dce_style = ASC_RET_USED_DCE_STYLE
    datagram = ASC_RET_DATAGRAM
    connection = ASC_RET_CONNECTION
    call_level = ASC_RET_CALL_LEVEL
    third_leg_failed = ASC_RET_THIRD_LEG_FAILED
    extended_error = ASC_RET_EXTENDED_ERROR
    stream = ASC_RET_STREAM
    integrity = ASC_RET_INTEGRITY
    licensing = ASC_RET_LICENSING
    identify = ASC_RET_IDENTIFY
    null_session = ASC_RET_NULL_SESSION
    allow_non_user_logons = ASC_RET_ALLOW_NON_USER_LOGONS
    allow_context_replay = ASC_RET_ALLOW_CONTEXT_REPLAY
    fragment_only = ASC_RET_FRAGMENT_ONLY
    no_token = ASC_RET_NO_TOKEN
    no_additional_token = ASC_RET_NO_ADDITIONAL_TOKEN


class ServerContextReq:
    delegate = ASC_REQ_DELEGATE
    mutual_auth = ASC_REQ_MUTUAL_AUTH
    replay_detect = ASC_REQ_REPLAY_DETECT
    sequence_detect = ASC_REQ_SEQUENCE_DETECT
    confidentiality = ASC_REQ_CONFIDENTIALITY
    use_session_key = ASC_REQ_USE_SESSION_KEY
    allocate_memory = ASC_REQ_ALLOCATE_MEMORY
    use_dce_style = ASC_REQ_USE_DCE_STYLE
    datagram = ASC_REQ_DATAGRAM
    connection = ASC_REQ_CONNECTION
    call_level = ASC_REQ_CALL_LEVEL
    fragment_supplied = ASC_REQ_FRAGMENT_SUPPLIED
    extended_error = ASC_REQ_EXTENDED_ERROR
    stream = ASC_REQ_STREAM
    integrity = ASC_REQ_INTEGRITY
    licensing = ASC_REQ_LICENSING
    identify = ASC_REQ_IDENTIFY
    allow_null_session = ASC_REQ_ALLOW_NULL_SESSION
    allow_non_user_logons = ASC_REQ_ALLOW_NON_USER_LOGONS
    allow_context_replay = ASC_REQ_ALLOW_CONTEXT_REPLAY
    fragment_to_fit = ASC_REQ_FRAGMENT_TO_FIT
    no_token = ASC_REQ_NO_TOKEN
    proxy_bindings = ASC_REQ_PROXY_BINDINGS
    allow_missing_binding = ASC_REQ_ALLOW_MISSING_BINDINGS


class SSPIQoP:
    wrap_no_encrypt = SECQOP_WRAP_NO_ENCRYPT


class TargetDataRep:
    native = SECURITY_NATIVE_DREP
    network = SECURITY_NETWORK_DREP


SecPkgAttrSizes = namedtuple('SecPkgAttrSizes', ['max_token', 'max_signature', 'block_size', 'security_trailer'])
SecPkgInfo = namedtuple('SecPkgInfo', ['capabilities', 'version', 'rpcid', 'max_token', 'name', 'comment'])


cdef class Credential:

    def __cinit__(Credential self):
        self.expiry = 0

    def __dealloc__(Credential self):
        if self.expiry:
            FreeCredentialsHandle(&self.handle)


cdef class SecurityContext:

    def __cinit__(SecurityContext self):
        self.context_attr = 0
        self.expiry = 0
        self.init = 0

    def __dealloc__(SecurityContext self):
        if self.init:
            DeleteSecurityContext(&self.handle)


cdef class SecBufferDesc:

    def __cinit__(SecBufferDesc self, list buffers not None, unsigned long version=SECBUFFER_VERSION):
        self._c_value = NativeSecBufferDesc(version, <unsigned long>len(buffers), NULL)
        self._buffers = buffers

        if buffers:
            self._c_value.pBuffers = <PSecBuffer>malloc(sizeof(NativeSecBuffer) * len(buffers))
            if not self._c_value.pBuffers:
                raise MemoryError("Cannot malloc SecBufferDesc buffer")

            for idx, buffer in enumerate(buffers):
                self[idx] = buffer

    cdef PSecBufferDesc __c_value__(SecBufferDesc self):
        return &self._c_value

    def __len__(SecBufferDesc self):
        return self._c_value.cBuffers

    def __getitem__(SecBufferDesc self, key):
        return self._buffers[key]

    def __setitem__(SecBufferDesc self, key, SecBuffer value not None):
        value.replace_ptr(&self._c_value.pBuffers[key])

    def __iter__(SecBufferDesc self):
        for val in self._buffers:
            yield val

    def __repr__(SecBufferDesc self):
        return "<{0}.{1}(ulVersion={2}, cBuffers={3})>".format(type(self).__module__, type(self).__name__,
            self._c_value.ulVersion, self._c_value.cBuffers)

    def __str__(SecBufferDesc self):
        return "{0}(ulVersion={1}, cBuffers={2})".format(type(self).__name__, self._c_value.ulVersion,
            self._c_value.cBuffers)

    @property
    def version(SecBufferDesc self):
        return self._c_value.ulVersion

    @version.setter
    def version(SecBufferDesc self, unsigned long value):
        self._c_value.ulVersion = value

    def __dealloc__(SecBufferDesc self):
        if self._c_value.pBuffers:
            free(self._c_value.pBuffers)
            self._c_value.pBuffers = NULL


cdef class SecBuffer:

    def __cinit__(SecBuffer self, unsigned long buffer_type, bytes buffer=None, unsigned long length=0):
        if buffer and length:
            raise ValueError("Only an empty buffer can be created with length")

        self.sys_alloc = 0
        self._is_owner = 1

        self.c_value = <PSecBuffer>calloc(1, sizeof(NativeSecBuffer))
        if not self.c_value:
            raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover

        self.buffer_type = buffer_type
        if buffer:
            self.buffer = buffer
        elif length:
            self._alloc_buffer(length)

    cdef replace_ptr(SecBuffer self, PSecBuffer ptr):
        # Copy the existing C buffer to the new pointer and free the memory this object was managing.
        memcpy(ptr, self.c_value, sizeof(NativeSecBuffer))
        free(self.c_value)

        # Make sure that this obj does not try and free the new pointer that was passed in (it is up to the caller to
        # do that).
        self._is_owner = 0
        self.c_value = ptr

    def __len__(SecBuffer self):
        return self.c_value.cbBuffer if self.c_value else 0

    def __repr__(SecBuffer self):
        return "<{0}.{1}(cbBuffer={2}, BufferType={3}, pvBuffer={4})>".format(
            type(self).__module__, type(self).__name__, self.c_value.cbBuffer, self.c_value.BufferType,
            repr(self.buffer))

    def __str__(SecBuffer self):
        return "{0}(cbBuffer={1}, BufferType={2}, pvBuffer={3!r})".format(type(self).__name__, self.c_value.cbBuffer,
            self.c_value.BufferType, self.buffer)

    @property
    def buffer_type(SecBuffer self):
        return self.c_value.BufferType

    @buffer_type.setter
    def buffer_type(SecBuffer self, unsigned long value):
        self.c_value.BufferType = value

    @property
    def buffer(SecBuffer self):
        if self.c_value.cbBuffer and self.c_value.pvBuffer:
            return (<char *>self.c_value.pvBuffer)[:self.c_value.cbBuffer]
        elif self.c_value.pvBuffer:
            return b""  # The size was 0 but pvBuffer was not a NULL pointer.
        else:
            return

    @buffer.setter
    def buffer(SecBuffer self, bytes value):
        value_len = len(value)
        self._alloc_buffer(value_len)
        memcpy(self.c_value.pvBuffer, <char *>value, value_len)

    def _alloc_buffer(SecBuffer self, unsigned long length):
        self._dealloc_buffer()

        # We store our allocated memory pointer so we know when we free we are freeing what we allocated.
        self._p_buffer = malloc(length)
        if not self._p_buffer:
            raise MemoryError("Cannot malloc SecBuffer buffer")  # pragma: no cover

        self.c_value.pvBuffer = self._p_buffer
        self.c_value.cbBuffer = length

    def _dealloc_buffer(SecBuffer self):
        # We do need to check if the actual C struct pvBuffer was allocated by Windows and call FreeContextBuffer on
        # that.
        if self.c_value != NULL and self.c_value.pvBuffer != NULL and self.sys_alloc:
            FreeContextBuffer(self.c_value.pvBuffer)

        if self.c_value:
            self.c_value.pvBuffer = NULL

        # Because the C struct pvBuffer may have a pointer set by Windows we track our allocated memory using an
        # internal attribute which we free().
        if self._p_buffer:
            free(self._p_buffer)
        self._p_buffer = NULL

    def __dealloc__(SecBuffer self):
        self._dealloc_buffer()

        if self.c_value != NULL and self._is_owner:
            free(self.c_value)

        self.c_value = NULL


cdef class _AuthIdentityBase:

    cdef void *__c_value__(_AuthIdentityBase self):
        return NULL


cdef class WinNTAuthIdentity(_AuthIdentityBase):

    def __cinit__(WinNTAuthIdentity self, unicode username, unicode domain, unicode password):
        self.c_value.Flags = SEC_WINNT_AUTH_IDENTITY_UNICODE
        self.username = username
        self.domain = domain
        self.password = password

    def __repr__(WinNTAuthIdentity self):
        domain = u"%s\\" % self.domain if self.domain else u""

        return to_native(u"<{0}.{1} {2}{3}>".format(type(self).__module__, type(self).__name__, domain,
            self.username or u""))

    def __str__(WinNTAuthIdentity self):
        domain = u"%s\\" % self.domain if self.domain else u""

        return to_native(u"{0}{1}".format(domain, self.username or u""))

    @property
    def username(WinNTAuthIdentity self):
        return self._username.to_text(self.c_value.UserLength)

    @username.setter
    def username(WinNTAuthIdentity self, unicode value):
        self._username = WideChar.from_text(value)
        self.c_value.User = <void *>self._username.buffer
        self.c_value.UserLength = max(self._username.length - 1, 0)

    @property
    def domain(WinNTAuthIdentity self):
        return self._domain.to_text(self.c_value.DomainLength)

    @domain.setter
    def domain(WinNTAuthIdentity self, unicode value):
        self._domain = WideChar.from_text(value)
        self.c_value.Domain = <void *>self._domain.buffer
        self.c_value.DomainLength = max(self._domain.length - 1, 0)

    @property
    def password(WinNTAuthIdentity self):
        return self._password.to_text(self.c_value.PasswordLength)

    @password.setter
    def password(WinNTAuthIdentity self, unicode value):
        self._password = WideChar.from_text(value)
        self.c_value.Password = <void *>self._password.buffer
        self.c_value.PasswordLength = max(self._password.length - 1, 0)

    cdef void *__c_value__(WinNTAuthIdentity self):
        return &self.c_value

    def __dealloc__(WinNTAuthIdentity self):
        self.username = None
        self.domain = None
        self.password = None


def accept_security_context(Credential credential not None, SecurityContext context not None,
    SecBufferDesc input_buffer=None, unsigned long context_req=0, unsigned long target_data_rep=SECURITY_NATIVE_DREP,
    SecBufferDesc output_buffer=None):

    cdef PCtxtHandle input_context = &context.handle if context.init else NULL
    cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
    cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
    cdef SECURITY_INTEGER expiry

    for buffer in (output_buffer or []):
        if len((<SecBuffer>buffer)) == 0:
            context_req |= ISC_REQ_ALLOCATE_MEMORY
            (<SecBuffer>buffer).sys_alloc = 1

    res = AcceptSecurityContext(&credential.handle, input_context, input, context_req, target_data_rep,
        &context.handle, output, &context.context_attr, &expiry)

    if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:
        res = CompleteAuthToken(&context.handle, output)

    if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
        PyErr_SetFromWindowsErr(res)

    context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
    context.init = 1

    return res


def acquire_credentials_handle(unicode principal, unicode package not None,
    unsigned long credential_use=SECPKG_CRED_OUTBOUND, _AuthIdentityBase auth_data=None):

    cdef WideChar w_principal = WideChar.from_text(principal)
    cdef WideChar w_package = WideChar.from_text(package)
    cdef void *p_auth_data = auth_data.__c_value__() if auth_data else NULL
    cdef Credential cred = Credential()
    cdef SECURITY_INTEGER expiry

    res = AcquireCredentialsHandleW(w_principal.buffer, w_package.buffer, credential_use, NULL, p_auth_data, NULL,
        NULL, &cred.handle, &expiry)

    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    cred.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart

    return cred


def decrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):
    cdef unsigned long qop = 0;

    res = DecryptMessage(&context.handle, message.__c_value__(), seq_no, &qop)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    return qop


def encrypt_message(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0,
    unsigned long qop=0):

    res = EncryptMessage(&context.handle, qop, message.__c_value__(), seq_no)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)


def initialize_security_context(Credential credential not None, SecurityContext context not None,
    unicode target_name not None, unsigned long context_req=0, SecBufferDesc input_buffer=None,
    unsigned long target_data_rep=SECURITY_NATIVE_DREP, SecBufferDesc output_buffer=None):

    cdef PCtxtHandle input_context = &context.handle if context.init else NULL
    cdef WideChar w_target_name = WideChar.from_text(target_name)
    cdef PSecBufferDesc input = input_buffer.__c_value__() if input_buffer else NULL
    cdef PSecBufferDesc output = output_buffer.__c_value__() if output_buffer else NULL
    cdef SECURITY_INTEGER expiry

    for buffer in (output_buffer or []):
        if len((<SecBuffer>buffer)) == 0:
            context_req |= ISC_REQ_ALLOCATE_MEMORY
            (<SecBuffer>buffer).sys_alloc = 1

    res = InitializeSecurityContextW(&credential.handle, input_context, w_target_name.buffer, context_req, 0,
        target_data_rep, input, 0, &context.handle, output, &context.context_attr, &expiry)

    if res in [_SEC_I_COMPLETE_AND_CONTINUE, _SEC_I_COMPLETE_NEEDED]:
        res = CompleteAuthToken(&context.handle, output)

    if res not in [_SEC_I_CONTINUE_NEEDED, _SEC_E_OK]:
        PyErr_SetFromWindowsErr(res)

    context.expiry = (<unsigned long long>expiry.HighPart << 32) | expiry.LowPart
    context.init = 1

    return res


def make_signature(SecurityContext context not None, unsigned long qop, SecBufferDesc message not None,
    unsigned long seq_no=0):

    res = MakeSignature(&context.handle, qop, message.__c_value__(), seq_no)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)


def query_context_attributes(SecurityContext context not None, unsigned long attribute):
    if attribute not in [SecPkgAttr.names, SecPkgAttr.package_info, SecPkgAttr.session_key, SecPkgAttr.sizes]:
        raise NotImplementedError("Only names, package_info, session_key, or sizes is implemented")

    if attribute == SecPkgAttr.names:
        return _query_context_names(context)

    elif attribute == SecPkgAttr.package_info:
        return _query_context_package_info(context)

    elif attribute == SecPkgAttr.session_key:
        return _query_context_session_key(context)

    else:
        return _query_context_sizes(context)


def _query_context_names(SecurityContext context not None):
    cdef SecPkgContext_Names info

    res = QueryContextAttributesW(&context.handle, SecPkgAttr.names, &info)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    try:
        return u16_to_text(info.sUserName, -1)
    finally:
        FreeContextBuffer(<void*>info.sUserName)

def _query_context_package_info(SecurityContext context not None):
    cdef SecPkgContext_PackageInfoW raw_info
    cdef PSecPkgInfoW info

    res = QueryContextAttributesW(&context.handle, SecPkgAttr.package_info, &raw_info)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    try:
        info = <PSecPkgInfoW>raw_info.PackageInfo
        return SecPkgInfo(info.fCapabilities, info.wVersion, info.wRPCID, info.cbMaxToken, u16_to_text(info.Name, -1),
            u16_to_text(info.Comment, -1))
    finally:
        FreeContextBuffer(<void*>raw_info.PackageInfo)


def _query_context_session_key(SecurityContext context not None):
    cdef SecPkgContext_SessionKey info

    res = QueryContextAttributesW(&context.handle, SecPkgAttr.session_key, &info)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    try:
        return (<char *>info.SessionKey)[:info.SessionKeyLength]
    finally:
        FreeContextBuffer(info.SessionKey)


def _query_context_sizes(SecurityContext context not None):
    cdef SecPkgContext_Sizes sizes

    res = QueryContextAttributesW(&context.handle, SecPkgAttr.sizes, &sizes)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    return SecPkgAttrSizes(sizes.cbMaxToken, sizes.cbMaxSignature, sizes.cbBlockSize, sizes.cbSecurityTrailer)


def verify_signature(SecurityContext context not None, SecBufferDesc message not None, unsigned long seq_no=0):
    cdef unsigned long qop = 0

    res = VerifySignature(&context.handle, message.__c_value__(), seq_no, &qop)
    if res != _SEC_E_OK:
        PyErr_SetFromWindowsErr(res)

    return qop



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type  # noqa (fixes E402 for the imports below)

import sys

PY2 = sys.version_info[0] == 2

if PY2:
    binary_type = str
    text_type = unicode
else:
    binary_type = bytes
    text_type = str


def _obj_str(obj, default):
    # First try to get the str() then repr() before falling back to the default value.
    for func in [str, repr]:
        try:
            obj = func(obj)
        except (UnicodeError, TypeError):
            continue
        else:
            return obj
    else:
        return default


def to_bytes(obj, encoding='utf-8', errors='strict', nonstring='str'):
    if isinstance(obj, binary_type):
        return obj
    elif isinstance(obj, text_type):
        return obj.encode(encoding, errors)

    if nonstring == 'str':
        return to_bytes(_obj_str(obj, b""), encoding=encoding, errors=errors)
    elif nonstring == 'passthru':
        return obj
    elif nonstring == 'empty':
        return b''
    else:
        raise ValueError("Invalid nonstring value '%s', expecting str, passthru, or empty" % nonstring)


def to_text(obj, encoding='utf-8', errors='strict', nonstring='str'):
    if isinstance(obj, text_type):
        return obj
    elif isinstance(obj, binary_type):
        return obj.decode(encoding, errors)

    if nonstring == 'str':
        try:
            obj = obj.__unicode__()
        except (AttributeError, UnicodeError):
            obj = _obj_str(obj, u"")

        return to_text(obj, errors=errors, encoding=encoding)
    elif nonstring == 'passthru':
        return obj
    elif nonstring == 'empty':
        return u''
    else:
        raise ValueError("Invalid nonstring value '%s', expecting repr, passthru, or empty" % nonstring)


if PY2:
    to_native = to_bytes
else:
    to_native = to_text



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type  # noqa (fixes E402 for the imports below)

import base64
import contextlib
import logging
import os
import sys
import tempfile

from spnego._compat import (
    Generator,
    List,
    Optional,
    Tuple,
    reraise,
)

from spnego._context import (
    ContextProxy,
    ContextReq,
    GSSMech,
    IOVWrapResult,
    IOVUnwrapResult,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
    wrap_system_error,
)

from spnego._text import (
    text_type,
    to_bytes,
    to_text,
)

from spnego.exceptions import (
    GSSError as NativeError,
    NegotiateOptions,
    SpnegoError,
)

from spnego.iov import (
    BufferType,
    IOVBuffer,
)


log = logging.getLogger(__name__)

HAS_GSSAPI = True
GSSAPI_IMP_ERR = None
try:
    import gssapi

    from gssapi.raw import (
        acquire_cred_with_password,
        ChannelBindings,
        exceptions as gss_errors,
        GSSError,
        inquire_sec_context_by_oid,
        set_sec_context_option,
    )
except ImportError:
    GSSAPI_IMP_ERR = sys.exc_info()
    HAS_GSSAPI = False
    log.debug("Python gssapi not available, cannot use any GSSAPIProxy protocols: %s" % str(GSSAPI_IMP_ERR[1]))


HAS_IOV = True
GSSAPI_IOV_IMP_ERR = None
try:
    from gssapi.raw import (
        IOV,
        IOVBufferType,
        unwrap_iov,
        wrap_iov,
    )
except ImportError as err:
    GSSAPI_IOV_IMP_ERR = sys.exc_info()
    HAS_IOV = False
    log.debug("Python gssapi IOV extension not available: %s" % str(GSSAPI_IOV_IMP_ERR[1]))

_GSS_C_INQ_SSPI_SESSION_KEY = "1.2.840.113554.1.2.2.5.5"

# https://github.com/simo5/gss-ntlmssp/blob/bfc7232dbb2259072a976fc9cdb6ae4bfd323304/src/gssapi_ntlmssp.h#L68
_GSS_NTLMSSP_RESET_CRYPTO_OID = '1.3.6.1.4.1.7165.655.1.3'

# https://github.com/krb5/krb5/blob/master/src/lib/gssapi/spnego/spnego_mech.c#L483
_GSS_SPNEGO_REQUIRE_MIC_OID_STRING = '1.3.6.1.4.1.7165.655.1.2'


def _available_protocols(options=None):  # type: (Optional[NegotiateOptions]) -> List[str, ...]
    """ Return a list of protocols that GSSAPIProxy can offer. """
    if not options:
        options = NegotiateOptions(0)

    protocols = []
    if HAS_GSSAPI:
        # We can't offer Kerberos if the caller requires WinRM wrapping and IOV isn't available.
        if not (options & NegotiateOptions.wrapping_winrm and not HAS_IOV):
            protocols = ['kerberos']

        # We can only offer NTLM if the mech is installed and can retrieve the functionality the caller desires.
        if _gss_ntlmssp_available(session_key=bool(options & NegotiateOptions.session_key)):
            protocols.append('ntlm')

        # We can only offer Negotiate if we can offer both Kerberos and NTLM.
        if len(protocols) == 2:
            protocols.append('negotiate')

    return protocols


def _create_iov_result(iov):  # type: (IOV) -> Tuple[IOVBuffer, ...]
    """ Converts GSSAPI IOV buffer to generic IOVBuffer result. """
    buffers = []
    for i in iov:
        buffer_entry = IOVBuffer(type=BufferType(i.type), data=i.value)
        buffers.append(buffer_entry)

    return tuple(buffers)


@contextlib.contextmanager
def _env_path(name, value, default_value):  # type: (str, str, str) -> Generator[None, None, None]
    """ Adds a value to a PATH-like env var and preserve the existing value if present. """
    orig_value = os.environ.get(name, None)
    os.environ[name] = '%s:%s' % (value, orig_value or default_value)
    try:
        yield

    finally:
        if orig_value:
            os.environ[name] = orig_value

        else:
            del os.environ[name]


@contextlib.contextmanager
def _krb5_conf(forwardable=False):  # type: (bool) -> Generator[None, None, None]
    """ Runs with a custom krb5.conf file that extends the existing config if present. """
    if forwardable:
        with tempfile.NamedTemporaryFile() as temp_cfg:
            temp_cfg.write(b"[libdefaults]\nforwardable = true\n")
            temp_cfg.flush()

            with _env_path('KRB5_CONFIG', temp_cfg.name, '/etc/krb5.conf'):
                yield

            return

    yield


def _get_gssapi_credential(mech, usage, username=None, password=None, context_req=None):
    # type: (gssapi.OID, str, Optional[text_type], Optional[text_type]) -> gssapi.creds.Credentials
    """Gets a set of credential(s).

    Will get a set of GSSAPI credential(s) for the mech specified. If the username and password is specified then a new
    set of credentials are explicitly required for the mech specified. Otherwise the credentials are retrieved by the
    cache as defined by the mech.

    The behaviour of this function is highly dependent on the GSSAPI implementation installed as well as what NTLM mech
    is available. Here are some of the known behaviours of each mech.

    Kerberos:
        Works for any GSSAPI implementation. The cache is the CCACHE which can be managed with `kinit`.

    NTLM:
        Only works with MIT krb5 and requires `gss-ntlmssp`_ to be installed. The cache that this mech uses is either
        a plaintext file specified by `NTLM_USER_FILE` in the format `DOMAIN:USERNAME:PASSWORD` or
        `:USER_UPN@REALM:PASSWORD` or it can be configured with winbind to a standalone Samba setup or with AD.

    SPNEGO:
        To work properly it requires both Kerberos and NTLM to be available where the latter only works with MIT krb5,
        see `NTLM` for more details. It attempts to get a credential for the all the mechs that SPNEGO supports so it
        will retrieve a Kerberos cred then NTLM.

    Args:
        mech: The mech OID to get the credentials for.
        usage: Either `initiate` for a client context or `accept` for a server context.
        username: The username to get the credentials for, if omitted then the default user is gotten from the cache.
        password: The password for the user, if omitted then the cached credentials is retrieved.

    Returns:
        gssapi.creds.Credentials: The credential set that was created/retrieved.

    .. _gss-ntlmssp:
        https://github.com/gssapi/gss-ntlmssp
    """
    if username:
        name_type = getattr(gssapi.NameType, 'user' if usage == 'initiate' else 'hostbased_service')
        username = gssapi.Name(base=username, name_type=name_type)

    if username and password:
        # NOTE: MIT krb5 < 1.14 would store this cred in the global cache but later versions used a private cache in
        # memory. There's not much we can do about this but document this behaviour and hope people upgrade to a newer
        # version.
        # GSSAPI offers no way to specify custom flags like forwardable. We use a temp conf file to ensure an explicit
        # cred with the delegate flag will actually be forwardable.
        forwardable = False
        forwardable_mechs = [gssapi.OID.from_int_seq(GSSMech.kerberos.value),
                             gssapi.OID.from_int_seq(GSSMech.spnego.value)]
        if context_req and context_req & ContextReq.delegate and mech in forwardable_mechs:
            forwardable = True

        with _krb5_conf(forwardable=forwardable):
            cred = acquire_cred_with_password(username, to_bytes(password), usage=usage, mechs=[mech])

        return cred.creds

    cred = gssapi.Credentials(name=username, usage=usage, mechs=[mech])

    # We don't need to check the actual lifetime, just trying to get the valid will have gssapi check the lifetime and
    # raise an ExpiredCredentialsError if it is expired.
    _ = cred.lifetime

    return cred


def _gss_ntlmssp_available(session_key=False):  # type: (bool) -> bool
    """Determine if NTLM is available through GSSAPI.

    NTLM support through GSSAPI is a complicated story. Because we rely on NTLM being available for SPNEGO fallback
    when Kerberos doesn't work we need to make sure the currently installed provider will give us what we need.

    Here is the current lay of the land for each GSSAPI provider.

    MIT KRB5:
        MIT KRB5 does not have NTLM builtin but it can be added with the `gss-ntlmssp`_ provider. We check to make sure
        the NTLM mech is installed and implements the required functions that are needed for SPNEGO support.

        The `gss-ntlmssp`_ provider only recently added support for retrieving its session key in v0.9.0. Not all
        callers need this behaviour so the `session_key` arg can be used to do a further check on that if needed.

    Heimdal:
        There are 2 major variants for Heimdal; 1. macOS' implementation, and 2. the actual Heimdal distribution. Each
        build has builtin "support" for NTLM but so far they are not usable for this library because:

        * macOS' implementation doesn't produce valid tokens, they are rejected by the server.
        * Pure Heimdal `gss_acquire_cred_with_password` isn't implemented for NTLM, no explicit creds.
        * Doesn't seem to produce a NTLM v2 message so the strength is even less than what our Python impl can offer.
        * It is doubtful it implements the required functions that MIT KRB5 relies on to get SPNEGO working.

        Because of these reasons we don't consider NTLM usable through GSSAPI on Heimdal based setups.

    Args:
        session_key: Whether the caller will want access to the session key of the context.

    Returns:
        bool: Whether NTLM is available to use (True) or not (False).

    .. _gss-ntlmssp:
        https://github.com/gssapi/gss-ntlmssp
    """
    # Cache the result so we don't run this check multiple times.
    try:
        res = _gss_ntlmssp_available.result
        return res['session_key'] if session_key else res['available']
    except AttributeError:
        pass

    ntlm_features = {
        'available': False,
        'session_key': False,
    }

    # If any of these calls results in a GSSError we treat that as NTLM being unusable because these are standard
    # behaviours we expect to work.
    ntlm = gssapi.OID.from_int_seq(GSSMech.ntlm.value)
    try:
        # This can be anything, the first NTLM message doesn't need a valid target name or credential.
        spn = gssapi.Name('http@test', name_type=gssapi.NameType.hostbased_service)
        cred = _get_gssapi_credential(ntlm, 'initiate', username='user', password='pass')
        context = gssapi.SecurityContext(creds=cred, usage='initiate', name=spn, mech=ntlm)

        context.step()  # Need to at least have a context set up before we can call gss_set_sec_context_option.

        # macOS' Heimdal implementation will work up to this point but the end messages aren't actually valid. Luckily
        # it does not implement 'GSS_NTLMSSP_RESET_CRYPTO_OID' so by running this we can weed out that broken impl.
        _gss_ntlmssp_reset_crypto(context)

        ntlm_features['available'] = True
    except GSSError as gss_err:
        log.debug("GSSAPI does not support required the NTLM interfaces: %s" % str(gss_err))
    else:
        # gss-ntlmssp only recently added support for GSS_C_INQ_SSPI_SESSION_KEY in v0.9.0, we check if it is present
        # before declaring session_key support is there as it might control whether it is used or not.
        # https://github.com/gssapi/gss-ntlmssp/issues/10
        try:
            inquire_sec_context_by_oid(context, gssapi.OID.from_int_seq(_GSS_C_INQ_SSPI_SESSION_KEY))
        except gss_errors.OperationUnavailableError as o_err:
            # (GSS_S_UNAVAILABLE | ERR_NOTAVAIL) is raised when ntlmssp does support GSS_C_INQ_SSPI_SESSION key but
            # the context is not yet established. Any other errors would mean this isn't supported and we can't use
            # the current version installed if we need session_key interrogation.
            # https://github.com/gssapi/gss-ntlmssp/blob/9d7a275a4d6494606fb54713876e4f5cbf4d1362/src/gss_sec_ctx.c#L1277
            if getattr(o_err, 'min_code', 0) == 1314127894:  # ERR_NOTAVAIL
                ntlm_features['session_key'] = True

            else:
                log.debug("GSSAPI ntlmssp does not support session key interrogation: %s" % str(o_err))

    _gss_ntlmssp_available.result = ntlm_features
    return _gss_ntlmssp_available(session_key=session_key)


def _gss_ntlmssp_reset_crypto(context, outgoing=True):  # type: (gssapi.SecurityContext, bool) -> None
    """ Resets the NTLM RC4 ciphers when being used with SPNEGO. """
    reset_crypto = gssapi.OID.from_int_seq(_GSS_NTLMSSP_RESET_CRYPTO_OID)
    value = b"\x00\x00\x00\x00" if outgoing else b"\x01\x00\x00\x00"
    set_sec_context_option(reset_crypto, context=context, value=value)


def _gss_sasl_description(mech):  # type: (gssapi.OID) -> Optional[bytes]
    """ Attempts to get the SASL description of the mech specified. """
    try:
        res = _gss_sasl_description.result
        return res[mech.dotted_form]

    except (AttributeError, KeyError):
        res = getattr(_gss_sasl_description, 'result', {})

    try:
        sasl_desc = gssapi.raw.inquire_saslname_for_mech(mech).mech_description
    except Exception as e:
        log.debug("gss_inquire_saslname_for_mech(%s) failed: %s" % (mech.dotted_form, str(e)))
        sasl_desc = None

    res[mech.dotted_form] = sasl_desc
    _gss_sasl_description.result = res
    return _gss_sasl_description(mech)


class GSSAPIProxy(ContextProxy):
    """GSSAPI proxy class for GSSAPI on Linux.

    This proxy class for GSSAPI exposes GSSAPI calls into a common interface for SPNEGO authentication. This context
    uses the Python gssapi library to interface with the gss_* calls to provider Kerberos, and potentially native
    ntlm/negotiate functionality.
    """
    def __init__(self, username=None, password=None, hostname=None, service=None, channel_bindings=None,
                 context_req=ContextReq.default, usage='initiate', protocol='negotiate', options=0, _is_wrapped=False):

        if not HAS_GSSAPI:
            reraise(ImportError("GSSAPIProxy requires the Python gssapi library"), GSSAPI_IMP_ERR)

        super(GSSAPIProxy, self).__init__(username, password, hostname, service, channel_bindings, context_req, usage,
                                          protocol, options, _is_wrapped)

        mech_str = {
            'kerberos': GSSMech.kerberos.value,
            'negotiate': GSSMech.spnego.value,
            'ntlm': GSSMech.ntlm.value,
        }[self.protocol]
        mech = gssapi.OID.from_int_seq(mech_str)

        cred = None
        try:
            cred = _get_gssapi_credential(mech, self.usage, username=username, password=password,
                                          context_req=context_req)
        except GSSError as gss_err:
            reraise(SpnegoError(base_error=gss_err, context_msg="Getting GSSAPI credential"))

        context_kwargs = {}

        if self.channel_bindings:
            context_kwargs['channel_bindings'] = ChannelBindings(
                initiator_address_type=self.channel_bindings.initiator_addrtype,
                initiator_address=self.channel_bindings.initiator_address,
                acceptor_address_type=self.channel_bindings.acceptor_addrtype,
                acceptor_address=self.channel_bindings.acceptor_address,
                application_data=self.channel_bindings.application_data
            )

        if self.usage == 'initiate':
            spn = to_text("%s@%s" % (service.lower() if service else 'host', hostname or 'unspecified'))
            context_kwargs['name'] = gssapi.Name(spn, name_type=gssapi.NameType.hostbased_service)
            context_kwargs['mech'] = mech
            context_kwargs['flags'] = self._context_req

        self._context = gssapi.SecurityContext(creds=cred, usage=self.usage, **context_kwargs)

    @classmethod
    def available_protocols(cls, options=None):
        return _available_protocols(options=options)

    @classmethod
    def iov_available(cls):
        # NOTE: Even if the IOV headers are unavailable, if NTLM was negotiated then IOV won't work. Unfortunately we
        # cannot determine that here as we may not know the protocol until after negotiation.
        return HAS_IOV

    @property
    def client_principal(self):
        if self.usage == 'accept':
            # Looks like a bug in python-gssapi where the value still has the terminating null char.
            return to_text(self._context.initiator_name).rstrip(u'\x00')

    @property
    def complete(self):
        return self._context.complete

    @property
    def negotiated_protocol(self):
        try:
            # For an acceptor this can be blank until the first token is received
            oid = self._context.mech.dotted_form
        except AttributeError:
            return

        return {
            GSSMech.kerberos.value: 'kerberos',
            GSSMech.ntlm.value: 'ntlm',

            # Only set until the negotiate process is complete, will change to one of the above once the context is
            # set up.
            GSSMech.spnego.value: 'negotiate',
        }.get(oid, 'unknown: %s' % self._context.mech.dotted_form)

    @property
    @wrap_system_error(NativeError, "Retrieving session key")
    def session_key(self):
        return inquire_sec_context_by_oid(self._context, gssapi.OID.from_int_seq(_GSS_C_INQ_SSPI_SESSION_KEY))[0]

    @wrap_system_error(NativeError, "Processing security token")
    def step(self, in_token=None):
        if not self._is_wrapped:
            log.debug("GSSAPI step input: %s", to_text(base64.b64encode(in_token or b"")))

        out_token = self._context.step(in_token)
        self._context_attr = int(self._context.actual_flags)

        if not self._is_wrapped:
            log.debug("GSSAPI step output: %s", to_text(base64.b64encode(out_token or b"")))

        return out_token

    @wrap_system_error(NativeError, "Wrapping data")
    def wrap(self, data, encrypt=True, qop=None):
        res = gssapi.raw.wrap(self._context, data, confidential=encrypt, qop=qop)

        # gss-ntlmssp used to hardcode the conf_state=0 which results in encrpted=False. Because we know it is always
        # sealed we just manually set to True.
        # https://github.com/gssapi/gss-ntlmssp/pull/15
        encrypted = True if self.negotiated_protocol == 'ntlm' else res.encrypted

        return WrapResult(data=res.message, encrypted=encrypted)

    @wrap_system_error(NativeError, "Wrapping IOV buffer")
    def wrap_iov(self, iov, encrypt=True, qop=None):
        iov_buffer = IOV(*self._build_iov_list(iov), std_layout=False)
        encrypted = wrap_iov(self._context, iov_buffer, confidential=encrypt, qop=qop)

        return IOVWrapResult(buffers=_create_iov_result(iov_buffer), encrypted=encrypted)

    def wrap_winrm(self, data):
        if self.negotiated_protocol == 'ntlm':
            # NTLM does not support IOV wrapping, luckily the header is a fixed size so we can split at that.
            wrap_result = self.wrap(data).data
            header = wrap_result[:16]
            enc_data = wrap_result[16:]
            padding = b""

        else:
            iov = self.wrap_iov([BufferType.header, data, BufferType.padding]).buffers
            header = iov[0].data
            enc_data = iov[1].data
            padding = iov[2].data or b""

        return WinRMWrapResult(header=header, data=enc_data + padding, padding_length=len(padding))

    @wrap_system_error(NativeError, "Unwrapping data")
    def unwrap(self, data):
        res = gssapi.raw.unwrap(self._context, data)

        # See wrap for more info.
        encrypted = True if self.negotiated_protocol == 'ntlm' else res.encrypted

        return UnwrapResult(data=res.message, encrypted=encrypted, qop=res.qop)

    @wrap_system_error(NativeError, "Unwrapping IOV buffer")
    def unwrap_iov(self, iov):
        iov_buffer = IOV(*self._build_iov_list(iov), std_layout=False)
        res = unwrap_iov(self._context, iov_buffer)

        return IOVUnwrapResult(buffers=_create_iov_result(iov_buffer), encrypted=res.encrypted, qop=res.qop)

    def unwrap_winrm(self, header, data):
        # This is an extremely weird setup, we need to use gss_unwrap for NTLM but for Kerberos it depends on the
        # underlying provider that is used. Right now the proper IOV buffers required to work on both AES and RC4
        # encrypted only works for MIT KRB5 whereas Heimdal fails. It currently mandates a padding buffer of a
        # variable size which we cannot achieve in the way that WinRM encrypts the data. This is fixed in the source
        # code but until it is widely distributed we just need to use a way that is known to just work with AES. To
        # ensure that MIT works on both RC4 and AES we check the description which differs between the 2 implemtations.
        # It's not perfect but I don't know of another way to achieve this until more time has passed.
        # https://github.com/heimdal/heimdal/issues/739
        sasl_desc = _gss_sasl_description(self._context.mech)

        # https://github.com/krb5/krb5/blob/f2e28f13156785851819fc74cae52100e0521690/src/lib/gssapi/krb5/gssapi_krb5.c#L686
        if sasl_desc and sasl_desc == b'Kerberos 5 GSS-API Mechanism':
            # TODO: Should done when self.negotiated_protocol == 'kerberos', above explains why this can't be done yet.
            iov = self.unwrap_iov([
                (IOVBufferType.header, header),
                data,
                IOVBufferType.data
            ]).buffers
            return iov[1].data

        else:
            return self.unwrap(header + data).data

    @wrap_system_error(NativeError, "Signing message")
    def sign(self, data, qop=None):
        return gssapi.raw.get_mic(self._context, data, qop=qop)

    @wrap_system_error(NativeError, "Verifying message")
    def verify(self, data, mic):
        return gssapi.raw.verify_mic(self._context, data, mic)

    @property
    def _context_attr_map(self):
        attr_map = [
            (ContextReq.delegate, 'delegate_to_peer'),
            (ContextReq.mutual_auth, 'mutual_authentication'),
            (ContextReq.replay_detect, 'replay_detection'),
            (ContextReq.sequence_detect, 'out_of_sequence_detection'),
            (ContextReq.confidentiality, 'confidentiality'),
            (ContextReq.integrity, 'integrity'),

            # Only present when the DCE extensions are installed.
            (ContextReq.identify, 'identify'),

            # Only present with newer versions of python-gssapi https://github.com/pythongssapi/python-gssapi/pull/218.
            (ContextReq.delegate_policy, 'ok_as_delegate'),
        ]
        attrs = []
        for spnego_flag, gssapi_name in attr_map:
            if hasattr(gssapi.RequirementFlag, gssapi_name):
                attrs.append((spnego_flag, getattr(gssapi.RequirementFlag, gssapi_name)))

        return attrs

    @property
    def _requires_mech_list_mic(self):
        try:
            require_mic = gssapi.OID.from_int_seq(_GSS_SPNEGO_REQUIRE_MIC_OID_STRING)
            res = inquire_sec_context_by_oid(self._context, require_mic)
        except GSSError:
            # Not all gssapi mechs implement this OID, the other mechListMIC rules still apply but are calc elsewhere.
            return False
        else:
            return b"\x01" in res

    def _convert_iov_buffer(self, iov_buffer):  # type: (IOVBuffer) -> Tuple[int, bool, Optional[bytes]]
        buffer_data = None
        buffer_alloc = False

        if isinstance(iov_buffer.data, bytes):
            buffer_data = iov_buffer.data
        elif isinstance(iov_buffer.data, int):
            # This shouldn't really occur on GSSAPI but is here to mirror what SSPI does.
            buffer_data = b"\x00" * iov_buffer.data
        else:
            auto_alloc = [BufferType.header, BufferType.padding, BufferType.trailer]

            buffer_alloc = iov_buffer.data
            if buffer_alloc is None:
                buffer_alloc = iov_buffer.type in auto_alloc

        return iov_buffer.type, buffer_alloc, buffer_data

    @wrap_system_error(NativeError, "NTLM reset crypto state")
    def _reset_ntlm_crypto_state(self, outgoing=True):
        if self.negotiated_protocol == 'ntlm':
            _gss_ntlmssp_reset_crypto(self._context, outgoing=outgoing)



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Kerberos Messages

Various Kerberos message structures that is used by 'pyspnego-parse.py' to unpack raw ASN.1 values and print out a
pretty structure for end users. This code is not used in the actual spnego authentication processes and is just used
for debugging purposes.
"""

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type  # noqa (fixes E402 for the imports below)

import base64
import collections
import datetime
import struct

from spnego._asn1 import (
    ASN1Value,
    extract_asn1_tlv,
    get_sequence_value,
    TagClass,
    unpack_asn1,
    unpack_asn1_bit_string,
    unpack_asn1_general_string,
    unpack_asn1_generalized_time,
    unpack_asn1_integer,
    unpack_asn1_octet_string,
    unpack_asn1_sequence,
    unpack_asn1_tagged_sequence,
)

from spnego._compat import (
    Dict,
    List,
    Optional,
    Union,

    add_metaclass,

    IntEnum,
    IntFlag,
)

from spnego._text import (
    text_type,
    to_text,
)


def _enum_labels(value, enum_type=None):  # type: (Union[int, IntEnum, IntFlag], Optional[type]) -> Dict[int, str]
    """ Gets the human friendly labels of a known enum and what value they map to. """
    def get_labels(v):
        return getattr(v, 'native_labels', lambda: {})()

    return get_labels(enum_type) if enum_type else get_labels(value)


def parse_enum(value, enum_type=None):  # type: (Union[int, IntEnum], Optional[type]) -> str
    """ Parses an IntEnum into a human representative object of that enum. """
    enum_name = 'UNKNOWN'

    labels = _enum_labels(value, enum_type)
    value = int(value) if isinstance(value, int) else value

    for v, name in labels.items():
        if value == v:
            enum_name = name
            break

    return "%s (%s)" % (enum_name, value)


def parse_flags(value, enum_type=None):  # type: (Union[int, IntFlag], Optional[type]) -> Dict[str, any]
    """ Parses an IntFlag into each flag value that is set. """
    raw_value = int(value)
    flags = []

    labels = _enum_labels(value, enum_type)
    value = int(value)

    for v, name in labels.items():
        if value & v == v:
            value &= ~v
            flags.append("%s (%d)" % (name, v))

    if value != 0:
        flags.append('UNKNOWN (%d)' % value)

    return {
        'raw': raw_value,
        'flags': flags,
    }


def parse_kerberos_token(token, secret=None, encoding=None):
    # type: (KerberosV5Msg, Optional[str], Optional[str]) -> Dict[str, any]
    """ Parses a KerberosV5Msg object to a dict. """
    if not encoding:
        encoding = 'utf-8'

    def parse_default(value):  # type: (any) -> any
        return value

    def parse_datetime(value):  # type: (datetime.datetime) -> text_type
        return to_text(value.isoformat())

    def parse_text(value):  # type: (bytes) -> text_type
        return to_text(value, encoding=encoding, errors='replace')

    def parse_bytes(value):  # type: (bytes) -> text_type
        return to_text(base64.b16encode(value))

    def parse_principal_name(value):  # type: (PrincipalName) -> Dict[str, any]
        return {
            'name-type': parse_enum(value.name_type),
            'name-string': [parse_text(v) for v in value.value],
        }

    def parse_host_address(value):  # type: (HostAddress) -> Dict[str, any]
        return {
            'addr-type': parse_enum(value.addr_type),
            'address': parse_text(value.value),
        }

    def parse_token(value):  # type: (any) -> Dict[str, any]
        return parse_kerberos_token(value, secret, encoding)

    if isinstance(token, bytes):
        return parse_bytes(token)

    msg = {}
    for name, attr_name, attr_type in token.PARSE_MAP:
        attr_value = getattr(token, attr_name)

        parse_args = []
        if isinstance(attr_type, tuple):
            parse_args.append(attr_type[1])
            attr_type = attr_type[0]

        parse_func = {
            ParseType.default: parse_default,
            ParseType.enum: parse_enum,
            ParseType.flags: parse_flags,
            ParseType.datetime: parse_datetime,
            ParseType.text: parse_text,
            ParseType.bytes: parse_bytes,
            ParseType.principal_name: parse_principal_name,
            ParseType.host_address: parse_host_address,
            ParseType.token: parse_token,
        }[attr_type]

        if attr_value is None:
            parsed_value = None

        elif isinstance(attr_value, list):
            parsed_value = [parse_func(v, *parse_args) if v is not None else None for v in attr_value]

        else:
            parsed_value = parse_func(attr_value, *parse_args)

        msg[name] = parsed_value

    return msg


def unpack_hostname(value):  # type: (ASN1Value) -> HostAddress
    """ Unpacks an ASN.1 value to a HostAddress. """
    s = unpack_asn1_tagged_sequence(value)

    name_type = KerberosHostAddressType(get_sequence_value(s, 0, 'HostAddress', 'addr-type', unpack_asn1_integer))
    name = get_sequence_value(s, 1, 'HostAddress', 'address', unpack_asn1_octet_string)

    return HostAddress(name_type, name)


def unpack_principal_name(value):  # type: (ASN1Value) -> PrincipalName
    """ Unpacks an ASN.1 value to a PrincipalName. """
    s = unpack_asn1_tagged_sequence(value)

    name_type = KerberosPrincipalNameType(get_sequence_value(s, 0, 'PrincipalName', 'name-type', unpack_asn1_integer))
    name = [unpack_asn1_general_string(n) for n in
            get_sequence_value(s, 1, 'PrincipalName', 'name-string', unpack_asn1_sequence)]

    return PrincipalName(name_type, name)


HostAddress = collections.namedtuple('HostAddress', ['addr_type', 'value'])
"""Kerberos HostAddress and HostAddresses

A Kerberos Host Address ASN.1 definition is defined in `RFC 4120 5.2.5`_::

    HostAddress     ::= SEQUENCE  {
        addr-type       [0] Int32,
        address         [1] OCTET STRING
    }

Attributes:
    addr_type (KerberosHostAddressType): The type of address that was encoded.
    value (bytes): The address as a byte string.

.. _RFC 4120 5.2.5:
    https://www.rfc-editor.org/rfc/rfc4120#section-5.2.5
"""

PrincipalName = collections.namedtuple('PrincipalName', ['name_type', 'value'])
"""Kerberos Realm and PrincipalName

A Kerberos Principal Name ASN.1 definition is defined in `RFC 4120 5.2.2`_::

    PrincipalName   ::= SEQUENCE {
        name-type       [0] Int32,
        name-string     [1] SEQUENCE OF KerberosString
    }

Attributes:
    name_type (KerberosPrincipalNameType): The type of name the was encoded.
    value (List[bytes]): Each component that forms the name.

.. _RFC 4120 5.2.2:
    https://www.rfc-editor.org/rfc/rfc4120#section-5.2.2
"""


# https://www.rfc-editor.org/rfc/rfc4120#section-5.5.1 - ap-options
class KerberosAPOptions(IntFlag):
    mutual_required = 0x00000020
    use_session_key = 0x00000040
    reserved = 0x00000080

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosAPOptions.mutual_required: 'mutual-required',
            KerberosAPOptions.use_session_key: 'use-session-key',
            KerberosAPOptions.reserved: 'reserved',
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1 - KDCOptions
class KerberosKDCOptions:  # TODO: Use IntEnum when we've dropped py27
    reserved = 0x80000000
    forwardable = 0x40000000
    forwarded = 0x20000000
    proxiable = 0x10000000
    proxy = 0x08000000
    allow_postdate = 0x04000000
    postdated = 0x02000000
    unused7 = 0x01000000
    renewable = 0x00800000
    unused9 = 0x00400000
    unused10 = 0x00200000
    opt_hardware_auth = 0x00100000
    unused12 = 0x00080000
    unused13 = 0x00040000
    constrained_delegation = 0x00020000
    canonicalize = 0x00010000
    request_anonymous = 0x00008000
    unused17 = 0x00004000
    unused18 = 0x00002000
    unused19 = 0x00001000
    unused20 = 0x00000800
    unused21 = 0x00000400
    unused22 = 0x00000200
    unused23 = 0x00000100
    unused24 = 0x00000080
    unused25 = 0x00000040
    disable_transited_check = 0x00000020
    renewable_ok = 0x00000010
    enc_tkt_in_skey = 0x00000008
    unused29 = 0x00000004
    renew = 0x00000002
    validate = 0x00000001

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosKDCOptions.reserved: 'reserved',
            KerberosKDCOptions.forwardable: 'forwardable',
            KerberosKDCOptions.forwarded: 'forwarded',
            KerberosKDCOptions.proxiable: 'proxiable',
            KerberosKDCOptions.proxy: 'proxy',
            KerberosKDCOptions.allow_postdate: 'allow-postdate',
            KerberosKDCOptions.postdated: 'postdated',
            KerberosKDCOptions.unused7: 'unused7',
            KerberosKDCOptions.renewable: 'renewable',
            KerberosKDCOptions.unused9: 'unused9',
            KerberosKDCOptions.unused10: 'unused10',
            KerberosKDCOptions.opt_hardware_auth: 'opt-hardware-auth',
            KerberosKDCOptions.unused12: 'unused12',
            KerberosKDCOptions.unused13: 'unused13',
            KerberosKDCOptions.constrained_delegation: 'constrained-delegation',
            KerberosKDCOptions.canonicalize: 'canonicalize',
            KerberosKDCOptions.request_anonymous: 'request-anonymous',
            KerberosKDCOptions.unused17: 'unused17',
            KerberosKDCOptions.unused18: 'unused18',
            KerberosKDCOptions.unused19: 'unused19',
            KerberosKDCOptions.unused20: 'unused20',
            KerberosKDCOptions.unused21: 'unused21',
            KerberosKDCOptions.unused22: 'unused22',
            KerberosKDCOptions.unused23: 'unused23',
            KerberosKDCOptions.unused24: 'unused24',
            KerberosKDCOptions.unused25: 'unused25',
            KerberosKDCOptions.disable_transited_check: 'disable-transited-check',
            KerberosKDCOptions.renewable_ok: 'renewable-ok',
            KerberosKDCOptions.enc_tkt_in_skey: 'enc-tkt-in-skey',
            KerberosKDCOptions.unused29: 'unused29',
            KerberosKDCOptions.renew: 'renew',
            KerberosKDCOptions.validate: 'validate',
        }


# https://ldapwiki.com/wiki/Kerberos%20Encryption%20Types - etypes
class KerberosEncryptionType(IntEnum):
    des_cbc_crc = 0x0001
    des_cbc_md4 = 0x0002
    des_cbc_md5 = 0x0003
    des_cbc_raw = 0x0004
    des3_cbc_raw = 0x0006
    des3_cbc_sha1 = 0x0010
    aes128_cts_hmac_sha1_96 = 0x0011
    aes256_cts_hmac_sha1_96 = 0x0012
    aes128_cts_hmac_sha256_128 = 0x0013
    aes256_cts_hmac_sha384_192 = 0x0014
    rc4_hmac = 0x0017
    rc4_hmac_exp = 0x0018
    camellia128_cts_cmac = 0x0019
    camellia256_cts_cmac = 0x001a

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosEncryptionType.des_cbc_crc: 'DES_CBC_CRC',
            KerberosEncryptionType.des_cbc_md4: 'DES_CBC_MD4',
            KerberosEncryptionType.des_cbc_md5: 'DES_CBC_MD5',
            KerberosEncryptionType.des_cbc_raw: 'DES_CBC_RAW',
            KerberosEncryptionType.des3_cbc_raw: 'DES3_CBC_RAW',
            KerberosEncryptionType.des3_cbc_sha1: 'DES3_CBC_SHA1',
            KerberosEncryptionType.aes128_cts_hmac_sha1_96: 'AES128_CTS_HMAC_SHA1_96',
            KerberosEncryptionType.aes256_cts_hmac_sha1_96: 'AES256_CTS_HMAC_SHA1_96',
            KerberosEncryptionType.aes128_cts_hmac_sha256_128: 'AES128_CTS_HMAC_SHA256_128',
            KerberosEncryptionType.aes256_cts_hmac_sha384_192: 'AES256_CTS_HMAC_SHA384_192',
            KerberosEncryptionType.rc4_hmac: 'RC4_HMAC',
            KerberosEncryptionType.rc4_hmac_exp: 'RC4_HMAC_EXP',
            KerberosEncryptionType.camellia128_cts_cmac: 'CAMELLIA128_CTS_CMAC',
            KerberosEncryptionType.camellia256_cts_cmac: 'CAMELLIA256_CTS_CMAC',
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-7.5.9
class KerberosErrorCode(IntEnum):
    none = 0
    name_exp = 1
    service_exp = 2
    bad_pvno = 3
    c_old_mast_kvno = 4
    s_old_mast_kvno = 5
    c_principal_unknown = 6
    s_principal_unknown = 7
    principal_not_unique = 8
    null_key = 9
    cannot_postdate = 10
    never_valid = 11
    policy = 12
    badoption = 13
    etype_nosupp = 14
    sumtype_nosupp = 15
    padata_type_nosupp = 16
    trtype_nosupp = 17
    client_revoked = 18
    service_revoked = 19
    tgt_revoked = 20
    client_notyet = 21
    service_notyet = 22
    key_expired = 23
    preauth_failed = 24
    preauth_required = 25
    server_nomatch = 26
    must_use_user2user = 27
    path_not_accepted = 28
    kdc_svc_unavailable = 29
    ap_bad_integrity = 31
    ap_txt_expired = 32
    ap_tkt_nyv = 33
    ap_repeat = 34
    ap_not_use = 35
    ap_badmatch = 36
    ap_skew = 37
    ap_badaddr = 38
    ap_badversion = 39
    ap_msg_type = 40
    ap_modified = 41
    ap_badorder = 42
    ap_badkeyver = 44
    ap_nokey = 45
    ap_mut_fail = 46
    ap_baddirection = 47
    ap_method = 48
    ap_badseq = 49
    ap_inapp_cksum = 50
    ap_path_not_accepted = 51
    response_too_big = 52
    generic = 60
    field_toolong = 61
    kdc_client_not_trusted = 62
    kdc_not_trusted = 53
    kdc_invalid_sig = 64
    kdc_key_too_weak = 65
    kdc_certificate_mismatch = 66
    ap_no_tgt = 67
    kdc_wrong_realm = 68
    ap_user_to_user_required = 69
    kdc_cant_verify_certificate = 70
    kdc_invalid_certificate = 71
    kdc_revoked_certificate = 72
    kdc_revocation_status_unknown = 73
    kdc_revocation_status_unavailable = 74
    kdc_client_name_mismatch = 75
    kdc_name_mismatch = 76

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosErrorCode.none: 'KDC_ERR_NONE',
            KerberosErrorCode.name_exp: 'KDC_ERR_NAME_EXP',
            KerberosErrorCode.service_exp: 'KDC_ERR_SERVICE_EXP',
            KerberosErrorCode.bad_pvno: 'KDC_ERR_BAD_PVNO',
            KerberosErrorCode.c_old_mast_kvno: 'KDC_ERR_C_OLD_MAST_KVNO',
            KerberosErrorCode.s_old_mast_kvno: 'KDC_ERR_S_OLD_MAST_KVNO',
            KerberosErrorCode.c_principal_unknown: 'KDC_ERR_C_PRINCIPAL_UNKNOWN',
            KerberosErrorCode.s_principal_unknown: 'KDC_ERR_S_PRINCIPAL_UNKNOWN',
            KerberosErrorCode.principal_not_unique: 'KDC_ERR_PRINCIPAL_NOT_UNIQUE',
            KerberosErrorCode.null_key: 'KDC_ERR_NULL_KEY',
            KerberosErrorCode.cannot_postdate: 'KDC_ERR_CANNOT_POSTDATE',
            KerberosErrorCode.never_valid: 'KDC_ERR_NEVER_VALID',
            KerberosErrorCode.policy: 'KDC_ERR_POLICY',
            KerberosErrorCode.badoption: 'KDC_ERR_BADOPTION',
            KerberosErrorCode.etype_nosupp: 'KDC_ERR_ETYPE_NOSUPP',
            KerberosErrorCode.sumtype_nosupp: 'KDC_ERR_SUMTYPE_NOSUPP',
            KerberosErrorCode.padata_type_nosupp: 'KDC_ERR_PADATA_TYPE_NOSUPP',
            KerberosErrorCode.trtype_nosupp: 'KDC_ERR_TRTYPE_NOSUPP',
            KerberosErrorCode.client_revoked: 'KDC_ERR_CLIENT_REVOKED',
            KerberosErrorCode.service_revoked: 'KDC_ERR_SERVICE_REVOKED',
            KerberosErrorCode.tgt_revoked: 'KDC_ERR_TGT_REVOKED',
            KerberosErrorCode.client_notyet: 'KDC_ERR_CLIENT_NOTYET',
            KerberosErrorCode.service_notyet: 'KDC_ERR_SERVICE_NOTYET',
            KerberosErrorCode.key_expired: 'KDC_ERR_KEY_EXPIRED',
            KerberosErrorCode.preauth_failed: 'KDC_ERR_PREAUTH_FAILED',
            KerberosErrorCode.preauth_required: 'KDC_ERR_PREAUTH_REQUIRED',
            KerberosErrorCode.server_nomatch: 'KDC_ERR_SERVER_NOMATCH',
            KerberosErrorCode.must_use_user2user: 'KDC_ERR_MUST_USE_USER2USER',
            KerberosErrorCode.path_not_accepted: 'KDC_ERR_PATH_NOT_ACCEPTED',
            KerberosErrorCode.kdc_svc_unavailable: 'KDC_ERR_SVC_UNAVAILABLE',
            KerberosErrorCode.ap_bad_integrity: 'KRB_AP_ERR_BAD_INTEGRITY',
            KerberosErrorCode.ap_txt_expired: 'KRB_AP_ERR_TKT_EXPIRED',
            KerberosErrorCode.ap_tkt_nyv: 'KRB_AP_ERR_TKT_NYV',
            KerberosErrorCode.ap_repeat: 'KRB_AP_ERR_REPEAT',
            KerberosErrorCode.ap_not_use: 'KRB_AP_ERR_NOT_US',
            KerberosErrorCode.ap_badmatch: 'KRB_AP_ERR_BADMATCH',
            KerberosErrorCode.ap_skew: 'KRB_AP_ERR_SKEW',
            KerberosErrorCode.ap_badaddr: 'KRB_AP_ERR_BADADDR',
            KerberosErrorCode.ap_badversion: 'KRB_AP_ERR_BADVERSION',
            KerberosErrorCode.ap_msg_type: 'KRB_AP_ERR_MSG_TYPE',
            KerberosErrorCode.ap_modified: 'KRB_AP_ERR_MODIFIED',
            KerberosErrorCode.ap_badorder: 'KRB_AP_ERR_BADORDER',
            KerberosErrorCode.ap_badkeyver: 'KRB_AP_ERR_BADKEYVER',
            KerberosErrorCode.ap_nokey: 'KRB_AP_ERR_NOKEY',
            KerberosErrorCode.ap_mut_fail: 'KRB_AP_ERR_MUT_FAIL',
            KerberosErrorCode.ap_baddirection: 'KRB_AP_ERR_BADDIRECTION',
            KerberosErrorCode.ap_method: 'KRB_AP_ERR_METHOD',
            KerberosErrorCode.ap_badseq: 'KRB_AP_ERR_BADSEQ',
            KerberosErrorCode.ap_inapp_cksum: 'KRB_AP_ERR_INAPP_CKSUM',
            KerberosErrorCode.ap_path_not_accepted: 'KRB_AP_PATH_NOT_ACCEPTED',
            KerberosErrorCode.response_too_big: 'KRB_ERR_RESPONSE_TOO_BIG',
            KerberosErrorCode.generic: 'KRB_ERR_GENERIC',
            KerberosErrorCode.field_toolong: 'KRB_ERR_FIELD_TOOLONG',
            KerberosErrorCode.kdc_client_not_trusted: 'KDC_ERROR_CLIENT_NOT_TRUSTED',
            KerberosErrorCode.kdc_not_trusted: 'KDC_ERROR_KDC_NOT_TRUSTED',
            KerberosErrorCode.kdc_invalid_sig: 'KDC_ERROR_INVALID_SIG',
            KerberosErrorCode.kdc_key_too_weak: 'KDC_ERR_KEY_TOO_WEAK',
            KerberosErrorCode.kdc_certificate_mismatch: 'KDC_ERR_CERTIFICATE_MISMATCH',
            KerberosErrorCode.ap_no_tgt: 'KRB_AP_ERR_NO_TGT',
            KerberosErrorCode.kdc_wrong_realm: 'KDC_ERR_WRONG_REALM',
            KerberosErrorCode.ap_user_to_user_required: 'KRB_AP_ERR_USER_TO_USER_REQUIRED',
            KerberosErrorCode.kdc_cant_verify_certificate: 'KDC_ERR_CANT_VERIFY_CERTIFICATE',
            KerberosErrorCode.kdc_invalid_certificate: 'KDC_ERR_INVALID_CERTIFICATE',
            KerberosErrorCode.kdc_revoked_certificate: 'KDC_ERR_REVOKED_CERTIFICATE',
            KerberosErrorCode.kdc_revocation_status_unknown: 'KDC_ERR_REVOCATION_STATUS_UNKNOWN',
            KerberosErrorCode.kdc_revocation_status_unavailable: 'KDC_ERR_REVOCATION_STATUS_UNAVAILABLE',
            KerberosErrorCode.kdc_client_name_mismatch: 'KDC_ERR_CLIENT_NAME_MISMATCH',
            KerberosErrorCode.kdc_name_mismatch: 'KDC_ERR_KDC_NAME_MISMATCH',
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-5.10
class KerberosMessageType(IntEnum):
    as_req = 10
    as_rep = 11
    tgs_req = 12
    tgs_rep = 13
    ap_req = 14
    ap_rep = 15
    error = 30

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosMessageType.as_req: 'AS-REQ',
            KerberosMessageType.as_rep: 'AS-REP',
            KerberosMessageType.tgs_req: 'TGS-REQ',
            KerberosMessageType.tgs_rep: 'TGS-REP',
            KerberosMessageType.ap_req: 'AP-REQ',
            KerberosMessageType.ap_rep: 'AP-REP',
            KerberosMessageType.error: 'KRB-ERROR',
        }


# https://www.iana.org/assignments/kerberos-parameters/kerberos-parameters.xhtml
class KerberosPADataType(IntEnum):
    tgs_req = 1
    enc_timestamp = 2
    pw_salt = 3
    reserved = 4
    enc_unix_time = 5
    sandia_secureid = 6
    sesame = 7
    osf_dce = 8
    cybersafe_secureid = 9
    afs3_salt = 10
    etype_info = 11
    sam_challenge = 12
    sam_response = 13
    pk_as_req_old = 14
    pk_as_rep_old = 15
    pk_as_req = 16
    pk_as_rep = 17
    pk_ocsp_response = 18
    etype_info2 = 19
    use_specified_kvno = 20
    svr_referral_info = 20
    sam_redirect = 21
    get_from_typed_data = 22
    td_padata = 22
    sam_etype_info = 23
    alt_princ = 24
    server_referral = 25
    sam_challenge2 = 30
    sam_response2 = 31
    extra_tgt = 41
    td_pkinit_cms_certificates = 101
    td_krb_principal = 102
    td_krb_realm = 103
    td_trusted_certifiers = 104
    td_certificate_index = 105
    td_app_defined_error = 106
    td_req_nonce = 107
    td_req_seq = 108
    td_dh_parameters = 109
    td_cms_digest_algorithms = 111
    td_cert_digest_algorithms = 112
    pac_request = 128
    for_user = 128
    for_x509_user = 130
    for_check_dups = 131
    as_checksum = 132
    fx_cookie = 133
    authentication_set = 134
    auth_set_selected = 165
    fx_fast = 136
    fx_error = 137
    encrypted_challenge = 138
    otp_challenge = 141
    otp_request = 142
    otp_confirm = 143
    otp_pin_change = 144
    epak_as_req = 145
    epak_as_rep = 146
    pkinit_kx = 147
    pku2u_name = 148
    enc_pa_rep = 149
    as_freshness = 150
    spake = 151
    kerb_key_list_req = 161
    kerb_key_list_rep = 162
    supported_etypes = 165
    extended_error = 166
    pac_options = 167

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosPADataType.tgs_req: 'PA-TGS-REQ',
            KerberosPADataType.enc_timestamp: 'PA-ENC-TIMESTAMP',
            KerberosPADataType.pw_salt: 'PA-PW-SALT',
            KerberosPADataType.reserved: 'reserved',
            KerberosPADataType.enc_unix_time: 'PA-ENC-UNIX-TIME',
            KerberosPADataType.sandia_secureid: 'PA-SANDIA-SECUREID',
            KerberosPADataType.sesame: 'PA-SESAME',
            KerberosPADataType.osf_dce: 'PA-OSF-DCE',
            KerberosPADataType.cybersafe_secureid: 'PA-CYBERSAFE-SECUREID',
            KerberosPADataType.afs3_salt: 'PA-AFS3-SALT',
            KerberosPADataType.etype_info: 'PA-ETYPE-INFO',
            KerberosPADataType.sam_challenge: 'PA-SAM-CHALLENGE',
            KerberosPADataType.sam_response: 'PA-SAM-RESPONSE',
            KerberosPADataType.pk_as_req_old: 'PA-PK-AS-REQ_OLD',
            KerberosPADataType.pk_as_rep_old: 'PA-PK-AS-REP_OLD',
            KerberosPADataType.pk_as_req: 'PA-PK-AS-REQ',
            KerberosPADataType.pk_as_rep: 'PA-PK-AS-REP',
            KerberosPADataType.pk_ocsp_response: 'PA-PK-OCSP-RESPONSE',
            KerberosPADataType.etype_info2: 'PA-ETYPE-INFO2',
            KerberosPADataType.use_specified_kvno: 'PA-USE-SPECIFIED-KVNO or PA-SVR-REFERRAL-INFO',
            KerberosPADataType.sam_redirect: 'PA-SAM-REDIRECT',
            KerberosPADataType.get_from_typed_data: 'PA-GET-FROM-TYPED-DATA',
            KerberosPADataType.td_padata: 'TD-PADATA',
            KerberosPADataType.sam_etype_info: 'PA-SAM-ETYPE-INFO',
            KerberosPADataType.alt_princ: 'PA-ALT-PRINC',
            KerberosPADataType.server_referral: 'PA-SERVER-REFERRAL',
            KerberosPADataType.sam_challenge2: 'PA-SAM-CHALLENGE2',
            KerberosPADataType.sam_response2: 'PA-SAM-RESPONSE2',
            KerberosPADataType.extra_tgt: 'PA-EXTRA-TGT',
            KerberosPADataType.td_pkinit_cms_certificates: 'TD-PKINIT-CMS-CERTIFICATES',
            KerberosPADataType.td_krb_principal: 'TD-KRB-PRINCIPAL',
            KerberosPADataType.td_krb_realm: 'TD-KRB-REALM',
            KerberosPADataType.td_trusted_certifiers: 'TD-TRUSTED-CERTIFIERS',
            KerberosPADataType.td_certificate_index: 'TD-CERTIFICATE-INDEX',
            KerberosPADataType.td_app_defined_error: 'TD-APP-DEFINED-ERROR',
            KerberosPADataType.td_req_nonce: 'TD-REQ-NONCE',
            KerberosPADataType.td_req_seq: 'TD-REQ-SEQ',
            KerberosPADataType.td_dh_parameters: 'TD_DH_PARAMETERS',
            KerberosPADataType.td_cms_digest_algorithms: 'TD-CMS-DIGEST-ALGORITHMS',
            KerberosPADataType.td_cert_digest_algorithms: 'TD-CERT-DIGEST-ALGORITHMS',
            KerberosPADataType.pac_request: 'PA-PAC-REQUEST',
            KerberosPADataType.for_user: 'PA-FOR_USER',
            KerberosPADataType.for_x509_user: 'PA-FOR-X509-USER',
            KerberosPADataType.for_check_dups: 'PA-FOR-CHECK_DUPS',
            KerberosPADataType.as_checksum: 'PA-AS-CHECKSUM',
            KerberosPADataType.fx_cookie: 'PA-FX-COOKIE',
            KerberosPADataType.authentication_set: 'PA-AUTHENTICATION-SET',
            KerberosPADataType.auth_set_selected: 'PA-AUTH-SET-SELECTED',
            KerberosPADataType.fx_fast: 'PA-FX-FAST',
            KerberosPADataType.fx_error: 'PA-FX-ERROR',
            KerberosPADataType.encrypted_challenge: 'PA-ENCRYPTED-CHALLENGE',
            KerberosPADataType.otp_challenge: 'PA-OTP-CHALLENGE',
            KerberosPADataType.otp_request: 'PA-OTP-REQUEST',
            KerberosPADataType.otp_confirm: 'PA-OTP-CONFIRM',
            KerberosPADataType.otp_pin_change: 'PA-OTP-PIN-CHANGE',
            KerberosPADataType.epak_as_req: 'PA-EPAK-AS-REQ',
            KerberosPADataType.epak_as_rep: 'PA-EPAK-AS-REP',
            KerberosPADataType.pkinit_kx: 'PA_PKINIT_KX',
            KerberosPADataType.pku2u_name: 'PA_PKU2U_NAME',
            KerberosPADataType.enc_pa_rep: 'PA-REQ-ENC-PA-REP',
            KerberosPADataType.as_freshness: 'PA_AS_FRESHNESS',
            KerberosPADataType.spake: 'PA-SPAKE',
            KerberosPADataType.kerb_key_list_req: 'KERB-KEY-LIST-REQ',
            KerberosPADataType.kerb_key_list_rep: 'KERB-KEY-LIST-REP',
            KerberosPADataType.supported_etypes: 'PA-SUPPORTED-ETYPES',
            KerberosPADataType.extended_error: 'PA-EXTENDED_ERROR',
            KerberosPADataType.pac_options: 'PA-PAC-OPTIONS',
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-6.2
class KerberosPrincipalNameType(IntEnum):
    unknown = 0
    principal = 1
    srv_inst = 2
    srv_hst = 3
    srv_xhst = 4
    uid = 5
    x500_principal = 6
    smtp_name = 7
    enterprise = 10

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosPrincipalNameType.unknown: 'NT-UNKNOWN',
            KerberosPrincipalNameType.principal: 'NT-PRINCIPAL',
            KerberosPrincipalNameType.srv_inst: 'NT-SRV-INST',
            KerberosPrincipalNameType.srv_hst: 'NT-SRV-HST',
            KerberosPrincipalNameType.srv_xhst: 'NT-SRV-XHST',
            KerberosPrincipalNameType.uid: 'NT-UID',
            KerberosPrincipalNameType.x500_principal: 'NT-X500-PRINCIPAL',
            KerberosPrincipalNameType.smtp_name: 'NT-SMTP-NAME',
            KerberosPrincipalNameType.enterprise: 'NT-ENTERPRISE',
        }


# https://www.rfc-editor.org/rfc/rfc4120#section-7.5.3
class KerberosHostAddressType(IntEnum):
    ipv4 = 2
    directional = 3
    chaos_net = 5
    xns = 6
    iso = 7
    decnet_phase_iv = 12
    apple_talk_ddp = 16
    netbios = 20
    ipv6 = 26

    @classmethod
    def native_labels(cls):  # type: () -> Dict[int, str]
        return {
            KerberosHostAddressType.ipv4: 'IPv4',
            KerberosHostAddressType.directional: 'Directional',
            KerberosHostAddressType.chaos_net: 'ChaosNet',
            KerberosHostAddressType.xns: 'XNS',
            KerberosHostAddressType.iso: 'ISO',
            KerberosHostAddressType.decnet_phase_iv: 'DECNET Phase IV',
            KerberosHostAddressType.apple_talk_ddp: 'AppleTalk DDP',
            KerberosHostAddressType.netbios: 'NetBios',
            KerberosHostAddressType.ipv6: 'IPv6',
        }


class ParseType(IntEnum):
    default = 0
    enum = 1
    flags = 2
    datetime = 3
    text = 4
    bytes = 5
    principal_name = 6
    host_address = 7
    token = 8


class _KerberosMsgType(type):
    __registry = {}

    def __init__(cls, name, bases, attributes):
        pvno = getattr(cls, 'PVNO', 0)

        if pvno not in cls.__registry:
            cls.__registry[pvno] = {}

        if hasattr(cls, 'MESSAGE_TYPE'):
            cls.__registry[pvno][cls.MESSAGE_TYPE] = cls

    def __call__(cls, sequence):
        # The KrbAsReq msg starts at 1 not 0, so do a check for that.
        pvno_idx = 0
        if 0 not in sequence:
            pvno_idx = 1

        pvno = unpack_asn1_integer(sequence[pvno_idx])
        message_type = unpack_asn1_integer(sequence[pvno_idx + 1])
        new_cls = cls.__registry[pvno].get(message_type, cls)
        return super(_KerberosMsgType, new_cls).__call__(sequence)


@add_metaclass(_KerberosMsgType)
class KerberosV5Msg:

    PVNO = 5

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        # This is only used if decoding an unknown Kerberos message type.
        self.sequence = sequence

    @staticmethod
    def unpack(value):  # type: (Union[bytes, ASN1Value]) -> KerberosV5Msg
        msg_sequence = unpack_asn1_tagged_sequence(value)
        return KerberosV5Msg(msg_sequence)


class KrbAsReq(KerberosV5Msg):
    """The KRB_AS_REQ message.

    The KRB_AS_REQ message is used when the client wishes to retrieve a the initial ticket for a service. The
    KRB_TGS_REQ message is identical except for the tag and msg-type is used when retrieving additional tickets for a
    service.

    The ASN.1 definition for the KDC-REQ structure is defined in `RFC 4120 5.4.1`_::

        KDC-REQ         ::= SEQUENCE {
            -- NOTE: first tag is [1], not [0]
            pvno            [1] INTEGER (5) ,
            msg-type        [2] INTEGER (10 -- AS -- | 12 -- TGS --),
            padata          [3] SEQUENCE OF PA-DATA OPTIONAL
                                -- NOTE: not empty --,
            req-body        [4] KDC-REQ-BODY
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        padata (PAData): The pre-authentication data.
        req_body (KdcReqBody): The body of the request.

    .. _RFC 4120 5.4.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1
    """
    MESSAGE_TYPE = KerberosMessageType.as_req

    PARSE_MAP = [
        ('pvno', 'PVNO', ParseType.default),
        ('msg-type', 'MESSAGE_TYPE', ParseType.enum),
        ('padata', 'padata', ParseType.token),
        ('req-body', 'req_body', ParseType.token)
    ]

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        def unpack_padata(value):
            return [PAData.unpack(p) for p in unpack_asn1_sequence(value)]

        self.padata = get_sequence_value(sequence, 3, 'KDC-REQ', 'pa-data', unpack_padata)
        self.req_body = get_sequence_value(sequence, 4, 'KDC-REQ', 'req-body', KdcReqBody.unpack)


class KrbAsRep(KerberosV5Msg):
    """The KRB_AS_REP message.

    The KRB_AS_REP message is used for a reply from the KDC to a KRB_AS_REQ message. The KRB_TGS_REP message is
    identical except for the tag and msg-type.

    The ASN.1 definition for the KDC-REP structure is defined in `RFC 4120 5.4.2`_::

        KDC-REP         ::= SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (11 -- AS -- | 13 -- TGS --),
            padata          [2] SEQUENCE OF PA-DATA OPTIONAL
                                -- NOTE: not empty --,
            crealm          [3] Realm,
            cname           [4] PrincipalName,
            ticket          [5] Ticket,
            enc-part        [6] EncryptedData
                                -- EncASRepPart or EncTGSRepPart,
                                -- as appropriate
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        padata (PAData): The pre-authentication data.
        crealm (bytes): The client realm.
        cname (PrincipalName): The client principal name.
        ticket (Ticket): The newly issued ticket.
        enc_part (EncryptedData): The encrypted part of the message.

    .. _RFC 4120 5.4.2:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.2
    """
    MESSAGE_TYPE = KerberosMessageType.as_rep

    PARSE_MAP = [
        ('pvno', 'PVNO', ParseType.default),
        ('msg-type', 'MESSAGE_TYPE', ParseType.enum),
        ('padata', 'padata', ParseType.token),
        ('crealm', 'crealm', ParseType.text),
        ('cname', 'cname', ParseType.principal_name),
        ('ticket', 'ticket', ParseType.token),
        ('enc-part', 'enc_part', ParseType.token),
    ]

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        def unpack_padata(value):
            return [PAData.unpack(p) for p in unpack_asn1_sequence(value)]

        self.padata = get_sequence_value(sequence, 2, 'KDC-REP', 'pa-data', unpack_padata)
        self.crealm = get_sequence_value(sequence, 3, 'KDC-REP', 'crealm', unpack_asn1_general_string)
        self.cname = get_sequence_value(sequence, 4, 'KDC-REP', 'cname', unpack_principal_name)
        self.ticket = get_sequence_value(sequence, 5, 'KDC-REP', 'ticket', Ticket.unpack)
        self.enc_part = get_sequence_value(sequence, 6, 'KDC-REP', 'enc-part', EncryptedData.unpack)


class KrbTgsReq(KrbAsReq):
    """ The KRB_TGS_REQ is the same as KRB_AS_REQ but with a different MESSAGE_TYPE. """
    MESSAGE_TYPE = KerberosMessageType.tgs_req


class KrbTgsRep(KrbAsRep):
    """ The KRB_TGS_REP is the same as KRB_AS_REP but with a different MESSAGE_TYPE. """
    MESSAGE_TYPE = KerberosMessageType.tgs_rep


class KrbApReq(KerberosV5Msg):
    """The KRB_AP_REQ message.

    The KRB_AP_REQ message contains is used to authenticate the initiator to an acceptor.

    The ASN.1 definition for the KRB_AP_REQ structure is defined in `RFC 4120 5.5.1`_::

        AP-REQ          ::= [APPLICATION 14] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (14),
            ap-options      [2] APOptions,
            ticket          [3] Ticket,
            authenticator   [4] EncryptedData -- Authenticator
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        ap_options (KerberosAPOptions): Options related to the AP request.
        ticket (Ticket): The ticket authenticating the client to the server.
        authenticator (EncryptedData): The encrypted authenticator.

    .. _RFC 4120 5.5.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.5.1
    """
    MESSAGE_TYPE = KerberosMessageType.ap_req

    PARSE_MAP = [
        ('pvno', 'PVNO', ParseType.default),
        ('msg-type', 'MESSAGE_TYPE', ParseType.enum),
        ('ap-options', 'ap_options', ParseType.flags),
        ('ticket', 'ticket', ParseType.token),
        ('authenticator', 'authenticator', ParseType.token),
    ]

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        raw_ap_options = get_sequence_value(sequence, 2, 'AP-REQ', 'ap-options', unpack_asn1_bit_string)
        ap_options = struct.unpack("<I", raw_ap_options)[0]

        try:
            ap_options = KerberosAPOptions(ap_options)
        except ValueError:
            pass

        self.ap_options = ap_options
        self.ticket = get_sequence_value(sequence, 3, 'AP-REQ', 'ticket', Ticket.unpack)
        self.authenticator = get_sequence_value(sequence, 4, 'AP-REQ', 'authenticator', EncryptedData.unpack)


class KrbApRep(KerberosV5Msg):
    """The KRB_AP_REP message.

    The KRB_AP_REP is a response to an application request `KRB_AP_REQ`.

    The ASN.1 definition for the KRB_AP_REP structure is defined in `RFC 4120 5.5.2`_::

        AP-REP          ::= [APPLICATION 15] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (15),
            enc-part        [2] EncryptedData -- EncAPRepPart
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        enc_part (EncryptedData): The encrypted authenticator.

    .. _RFC 4120 5.5.2:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.5.2
    """
    MESSAGE_TYPE = KerberosMessageType.ap_rep

    PARSE_MAP = [
        ('pvno', 'PVNO', ParseType.default),
        ('msg-type', 'MESSAGE_TYPE', ParseType.enum),
        ('enc-part', 'enc_part', ParseType.token),
    ]

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        self.enc_part = get_sequence_value(sequence, 2, 'AP-REP', 'enc-part', EncryptedData.unpack)


class KrbError(KerberosV5Msg):
    """The KRB_ERROR message.

    The KRB_ERROR is a message sent in the occurrence of an error.

    The ASN.1 definition for the KRB_ERROR structure is defined in `RFC 4120 5.9.1`_::

        KRB-ERROR       ::= [APPLICATION 30] SEQUENCE {
            pvno            [0] INTEGER (5),
            msg-type        [1] INTEGER (30),
            ctime           [2] KerberosTime OPTIONAL,
            cusec           [3] Microseconds OPTIONAL,
            stime           [4] KerberosTime,
            susec           [5] Microseconds,
            error-code      [6] Int32,
            crealm          [7] Realm OPTIONAL,
            cname           [8] PrincipalName OPTIONAL,
            realm           [9] Realm -- service realm --,
            sname           [10] PrincipalName -- service name --,
            e-text          [11] KerberosString OPTIONAL,
            e-data          [12] OCTET STRING OPTIONAL
        }

    Args:
        sequence: The ASN.1 sequence value as a dict to unpack.

    Attributes:
        ctime (datetime.datetime): The current time on the client's host.
        cusec (int): The microsecond part of the client's timestamp.
        stime (datetime.datetime): The current time of the server.
        susec (int): The microsecond part of the server's timestamp.
        error_code (KerberosErrorCode): THe error code returned by the kerberos when a request fails.
        crealm (bytes): The realm that issues a ticket.
        cname (PrincipalName): The principal name in the ticket.
        realm (bytes): The service realm.
        sname (PrincipalName): The service name.
        e_text (bytes): Additional text to explain the error code.
        e_data (bytes): Additional data about the error.

    .. _RFC 4120 5.9.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.9.1
    """
    MESSAGE_TYPE = KerberosMessageType.error

    PARSE_MAP = [
        ('pvno', 'PVNO', ParseType.default),
        ('msg-type', 'MESSAGE_TYPE', ParseType.enum),
        ('ctime', 'ctime', ParseType.datetime),
        ('cusec', 'cusec', ParseType.default),
        ('stime', 'stime', ParseType.datetime),
        ('susec', 'susec', ParseType.default),
        ('error-code', 'error_code', ParseType.enum),
        ('crealm', 'crealm', ParseType.text),
        ('cname', 'cname', ParseType.principal_name),
        ('realm', 'realm', ParseType.text),
        ('sname', 'sname', ParseType.principal_name),
        ('e-text', 'e_text', ParseType.text),
        ('e-data', 'e_data', ParseType.bytes),
    ]

    def __init__(self, sequence):  # type: (Dict[int, ASN1Value]) -> None
        self.ctime = get_sequence_value(sequence, 2, 'KRB-ERROR', 'ctime', unpack_asn1_generalized_time)
        self.cusec = get_sequence_value(sequence, 3, 'KRB-ERROR', 'cusec', unpack_asn1_integer)
        self.stime = get_sequence_value(sequence, 4, 'KRB-ERROR', 'stime', unpack_asn1_generalized_time)
        self.susec = get_sequence_value(sequence, 5, 'KRB-ERROR', 'susec', unpack_asn1_integer)
        self.error_code = KerberosErrorCode(get_sequence_value(sequence, 6, 'KRB-ERROR', 'error-code',
                                                               unpack_asn1_integer))
        self.crealm = get_sequence_value(sequence, 7, 'KRB-ERROR', 'crealm', unpack_asn1_general_string)
        self.cname = get_sequence_value(sequence, 8, 'KRB-ERROR', 'cname', unpack_principal_name)
        self.realm = get_sequence_value(sequence, 9, 'KRB-ERROR', 'realm', unpack_asn1_general_string)
        self.sname = get_sequence_value(sequence, 10, 'KRB-ERROR', 'realm', unpack_principal_name)
        self.e_text = get_sequence_value(sequence, 11, 'KRB-ERROR', 'e-text', unpack_asn1_general_string)
        self.e_data = get_sequence_value(sequence, 12, 'KRB-ERROR', 'e-data', unpack_asn1_octet_string)


class PAData:
    """Kerberos PA-DATA.

    The ASN.1 definition for the PA-DATA structure is defined in `RFC 4120 5.2.7`_::

        PA-DATA         ::= SEQUENCE {
            -- NOTE: first tag is [1], not [0]
            padata-type     [1] Int32,
            padata-value    [2] OCTET STRING -- might be encoded AP-REQ
        }

    Args:
        data_type: Indicates the type of data the value represents.
        value: The PAData value, usually the DER encoding of another message.

    Attributes:
        data_type (Union[int, KerberosPADataType]): See args.
        b_value (bytes): The raw bytes of padata-value, use `value` to get a structured object of these bytes if
            available.

    .. RFC 4120 5.2.7:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.7
    """
    PARSE_MAP = [
        ('padata-type', 'data_type', ParseType.enum),
        ('padata-value', 'value', ParseType.token),
    ]

    def __init__(self, data_type, value):  # type: (Union[int, KerberosPADataType], bytes) -> None
        self.data_type = data_type
        self.b_value = value

    @property
    def value(self):
        if self.data_type == KerberosPADataType.tgs_req:
            # Special edge case for this PA type due to how the messages require the data in unpack
            return KrbTgsReq.unpack(unpack_asn1(unpack_asn1(self.b_value)[0].b_data)[0])

        unpack_func = {
            KerberosPADataType.enc_timestamp: (EncryptedData.unpack, False),
            KerberosPADataType.etype_info2: (PAETypeInfo2.unpack, True),
        }.get(self.data_type, None)

        if unpack_func:
            b_value = unpack_asn1(self.b_value)[0]
            if unpack_func[1]:
                # Is a SEQUENCE OF (list of entries).
                return [unpack_func[0](v) for v in unpack_asn1_sequence(b_value)]

            else:
                return unpack_func[0](b_value.b_data)

        else:
            return self.b_value

    @staticmethod
    def unpack(value):  # type: (ASN1Value) -> PAData
        sequence = unpack_asn1_tagged_sequence(value)

        def unpack_data_type(value):
            int_val = unpack_asn1_integer(value)

            try:
                return KerberosPADataType(int_val)
            except ValueError:
                return int_val

        data_type = get_sequence_value(sequence, 1, 'PA-DATA', 'padata-type', unpack_data_type)
        value = get_sequence_value(sequence, 2, 'PA-DATA', 'padata-value', unpack_asn1_octet_string)

        return PAData(data_type, value)


class PAETypeInfo2:
    """Kerberos PA-ETYPE-INFO2 container.

    The ASN.1 definition for the PA-ETYPE-INFO2 structure is defined in `RFC 4120 5.2.7.5`_::

        ETYPE-INFO2-ENTRY       ::= SEQUENCE {
            etype           [0] Int32,
            salt            [1] KerberosString OPTIONAL,
            s2kparams       [2] OCTET STRING OPTIONAL
        }

    Args:
        etype: The etype that defines the cipher used.
        salt: The used in the cipher associated with the cryptosystem.
        s2kparams: Extra params to be interpreted by the cipher associated with the cryptosystem.

    Attributes:
        etype (KerberosEncryptionType): See args.
        salt (Optional[bytes]): See args.
        s2kparams (Optional[bytes]): See args.

    .. RFC 4120 5.2.7.5:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.7.5
    """
    PARSE_MAP = [
        ('etype', 'etype', ParseType.enum),
        ('salt', 'salt', ParseType.bytes),
        ('s2kparams', 's2kparams', ParseType.bytes),
    ]

    def __init__(self, etype, salt, s2kparams):
        # type: (KerberosEncryptionType, Optional[bytes], Optional[bytes]) -> None
        self.etype = etype
        self.salt = salt
        self.s2kparams = s2kparams

    @staticmethod
    def unpack(value):
        sequence = unpack_asn1_tagged_sequence(value)

        etype = KerberosEncryptionType(get_sequence_value(sequence, 0, 'PA-ETYPE-INFO2', 'etype', unpack_asn1_integer))
        salt = get_sequence_value(sequence, 1, 'ETYPE-INFO2-ENTRY', 'salt', unpack_asn1_general_string)
        s2kparams = get_sequence_value(sequence, 2, 'ETYPE-INFO2-ENTRY', 's2kparams', unpack_asn1_octet_string)

        return PAETypeInfo2(etype, salt, s2kparams)


class EncryptedData:
    """Kerberos EncryptedData container.

    The ASN.1 definition for the EncryptedData structure is defined in `RFC 4120 5.2.9`_::

        EncryptedData   ::= SEQUENCE {
            etype   [0] Int32 -- EncryptionType --,
            kvno    [1] UInt32 OPTIONAL,
            cipher  [2] OCTET STRING -- ciphertext
        }

    Args:
        etype: The encryption algorithm that was used to encipher the cipher.
        kvno: The version number of the key under which data is encrypted. It is only present in messages encrypted
            under long lasting keys.
        cipher: The enciphered text.

    Attributes:
        etype (KerberosEncryptionType): See args.
        kvno (Optional[int]): See args.
        cipher (bytes): See args.

    .. RFC 4120 5.2.9:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.2.9
    """
    PARSE_MAP = [
        ('etype', 'etype', ParseType.enum),
        ('kvno', 'kvno', ParseType.default),
        ('cipher', 'cipher', ParseType.bytes),
    ]

    def __init__(self, etype, kvno, cipher):  # type: (KerberosEncryptionType, Optional[int], bytes) -> None
        self.etype = etype
        self.kvno = kvno
        self.cipher = cipher

    @staticmethod
    def unpack(value):
        sequence = unpack_asn1_tagged_sequence(value)

        etype = KerberosEncryptionType(get_sequence_value(sequence, 0, 'EncryptedData', 'etype', unpack_asn1_integer))
        kvno = get_sequence_value(sequence, 1, 'EncryptedData', 'kvno', unpack_asn1_integer)
        cipher = get_sequence_value(sequence, 2, 'EncryptedData', 'cipher', unpack_asn1_octet_string)

        return EncryptedData(etype, kvno, cipher)


class Ticket:
    """Kerberos Ticket.

    The ASN.1 definition for the Ticket structure is defined in `RFC 4120 5.3`_::

        Ticket          ::= [APPLICATION 1] SEQUENCE {
            tkt-vno         [0] INTEGER (5),
            realm           [1] Realm,
            sname           [2] PrincipalName,
            enc-part        [3] EncryptedData -- EncTicketPart
        }

    Args:
        tkt_vno: The version number for the ticket format.
        realm: The realm that issued a ticket.
        sname: All the name components of the server's identity.
        enc_part: The encrypted part of the ticket, it is encrypted in the key shared by Kerberos and the end server.

    Attributes:
        tkt_vno (int): See args.
        realm (bytes): See args.
        sname (List[PrincipalName]): See args.
        enc_part (EncryptedData): See args.

    .. _RFC 4120 5.3:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.3
    """
    PARSE_MAP = [
        ('tkt-vno', 'tkt_vno', ParseType.default),
        ('realm', 'realm', ParseType.text),
        ('sname', 'sname', ParseType.principal_name),
        ('enc-part', 'enc_part', ParseType.token),
    ]

    def __init__(self, tkt_vno, realm, sname, enc_part):
        # type; (int, bytes, List[PrincipalName], KrbEncryptedData) -> None
        self.tkt_vno = tkt_vno
        self.realm = realm
        self.sname = sname
        self.enc_part = enc_part

    @staticmethod
    def unpack(value):  # type: (Union[bytes, ASN1Value]) -> Ticket
        b_data = extract_asn1_tlv(value, TagClass.application, 1)
        sequence = unpack_asn1_tagged_sequence(unpack_asn1(b_data)[0])

        tkt_vno = get_sequence_value(sequence, 0, 'Ticket', 'tkt-vno', unpack_asn1_integer)
        realm = get_sequence_value(sequence, 1, 'Ticket', 'realm', unpack_asn1_general_string)
        sname = get_sequence_value(sequence, 2, 'Ticket', 'sname', unpack_principal_name)
        enc_part = get_sequence_value(sequence, 3, 'Ticket', 'enc-part', EncryptedData.unpack)

        return Ticket(tkt_vno, realm, sname, enc_part)


class KdcReqBody:
    """The KRB_AS_REQ message.

    The KRB_AS_REQ message is used when the client wishes to retrieve a the initial ticket for a service. The
    KRB_TGS_REQ message is identical except for the tag and msg-type is used when retrieving additional tickets for a
    service.

    The ASN.1 definition for the KDC-REQ structure is defined in `RFC 4120 5.4.1`_::

        KDC-REQ-BODY    ::= SEQUENCE {
            kdc-options             [0] KDCOptions,
            cname                   [1] PrincipalName OPTIONAL
                                        -- Used only in AS-REQ --,
            realm                   [2] Realm
                                        -- Server's realm
                                        -- Also client's in AS-REQ --,
            sname                   [3] PrincipalName OPTIONAL,
            from                    [4] KerberosTime OPTIONAL,
            till                    [5] KerberosTime,
            rtime                   [6] KerberosTime OPTIONAL,
            nonce                   [7] UInt32,
            etype                   [8] SEQUENCE OF Int32 -- EncryptionType
                                        -- in preference order --,
            addresses               [9] HostAddresses OPTIONAL,
            enc-authorization-data  [10] EncryptedData OPTIONAL
                                        -- AuthorizationData --,
            additional-tickets      [11] SEQUENCE OF Ticket OPTIONAL
                                        -- NOTE: not empty
        }

    Args:
        kdc_options: Flags desired by the client and other behaviour desired.
        cname: The client name.
        realm: The realm part of the server's principal.
        sname: The service name.
        postdated_from: When the requested ticket is to be posted from.
        postdated_till: The expiration date requested by the client.
        rtime: The requested renew-till time.
        nonce: Random number generated by the client.
        etype: The desired encryption algorithm to be used in priority order.
        addresses: Addresses from which the requested ticket is to be valid.
        enc_authorization_data: Encrypted authorization data.
        additional_tickets: Additional tickets to be optionally included in a request.

    Attributes:
        kdc_options (int): See args.
        cname (Optional[PrincipalName]): See args.
        realm (bytes): See args.
        sname (PrincipalName): See args.
        postdated_from (Optional[datetime.datetime]): See args.
        postdated_till (datetime.datetime): See args.
        rtime (Optional[datetime.datetime]): See args.
        nonce (int): See args.
        etype (List[KerberosEncryptionType]): See args.
        addresses (Optional[List[HostAddress]]): See args.
        enc_authorization_data (Optional[EncryptedData]): See args.
        additional_tickets (Optional[Ticket]): See args.

    .. _RFC 4120 5.4.1:
        https://www.rfc-editor.org/rfc/rfc4120#section-5.4.1
    """
    PARSE_MAP = [
        ('kdc-options', 'kdc_options', (ParseType.flags, KerberosKDCOptions)),
        ('cname', 'cname', ParseType.principal_name),
        ('realm', 'realm', ParseType.text),
        ('sname', 'sname', ParseType.principal_name),
        ('from', 'postdated_from', ParseType.datetime),
        ('till', 'postdated_till', ParseType.datetime),
        ('rtime', 'rtime', ParseType.datetime),
        ('nonce', 'nonce', ParseType.default),
        ('etype', 'etype', ParseType.enum),
        ('addresses', 'addresses', ParseType.host_address),
        ('enc-authorization-data', 'enc_authorization_data', ParseType.token),
        ('additional-tickets', 'additional_tickets', ParseType.token),
    ]

    def __init__(self, kdc_options, cname, realm, sname, postdated_from, postdated_till, rtime, nonce, etype,
                 addresses, enc_authorization_data, additional_tickets):
        # type: (int, Optional[PrincipalName], bytes, PrincipalName, Optional[datetime.datetime], datetime.datetime, Optional[datetime.datetime], int, List[KerberosEncryptionType], Optional[List[HostAddress]], Optional[EncryptedData], Optional[List[Ticket]]) -> None  # noqa
        self.kdc_options = kdc_options
        self.cname = cname
        self.realm = realm
        self.sname = sname
        self.postdated_from = postdated_from
        self.postdated_till = postdated_till
        self.rtime = rtime
        self.nonce = nonce
        self.etype = etype
        self.addresses = addresses
        self.enc_authorization_data = enc_authorization_data
        self.additional_tickets = additional_tickets

    @staticmethod
    def unpack(value):  # type: (Union[bytes, ASN1Value]) -> KdcReqBody
        sequence = unpack_asn1_tagged_sequence(value)

        def unpack_kdc_options(value):
            b_data = unpack_asn1_bit_string(value)
            return struct.unpack(">I", b_data)[0]

        def unpack_etype(value):
            return [KerberosEncryptionType(unpack_asn1_integer(e)) for e in unpack_asn1_sequence(value)]

        def unpack_addresses(value):
            return [unpack_hostname(h) for h in unpack_asn1_sequence(value)]

        def unpack_ticket(value):
            return [Ticket.unpack(t) for t in unpack_asn1_sequence(value)]

        kdc_options = get_sequence_value(sequence, 0, 'KDC-REQ-BODY', 'kdc-options', unpack_kdc_options)
        cname = get_sequence_value(sequence, 1, 'KDC-REQ-BODY', 'cname', unpack_principal_name)
        realm = get_sequence_value(sequence, 2, 'KDC-REQ-BODY', 'realm', unpack_asn1_general_string)
        sname = get_sequence_value(sequence, 3, 'KDC-REQ-BODY', 'sname', unpack_principal_name)
        postdated_from = get_sequence_value(sequence, 4, 'KDC-REQ-BODY', 'from', unpack_asn1_generalized_time)
        postdated_till = get_sequence_value(sequence, 5, 'KDC-REQ-BODY', 'till', unpack_asn1_generalized_time)
        rtime = get_sequence_value(sequence, 6, 'KDC-REQ-BODY', 'rtime', unpack_asn1_generalized_time)
        nonce = get_sequence_value(sequence, 7, 'KDC-REQ-BODY', 'nonce', unpack_asn1_integer)
        etype = get_sequence_value(sequence, 8, 'KDC-REQ-BODY', 'etype', unpack_etype)
        addresses = get_sequence_value(sequence, 9, 'KDC-REQ-BODY', 'addresses', unpack_addresses)
        enc_auth_data = get_sequence_value(sequence, 10, 'KDC-REQ-BODY', 'enc-authorization-data',
                                           EncryptedData.unpack)
        additional_tickets = get_sequence_value(sequence, 11, 'KDC-REQ-BODY', 'additional-tickets', unpack_ticket)

        return KdcReqBody(kdc_options, cname, realm, sname, postdated_from, postdated_till, rtime, nonce, etype,
                          addresses, enc_auth_data, additional_tickets)
